import asyncio
import datetime
import os
from pathlib import Path
from typing import Optional

import aiohttp
from loguru import logger
from pydantic import BaseModel, validator

from ..areader import AsyncBaseReader
from ..paper_with_image import Paper
from ..provider import async_arxiv as arxiv


class PaperParams(BaseModel):
    pdf: str
    query: str
    key_word: str
    filter_keys: Optional[list[str]] = None
    max_results: int
    sort: str
    save_image: bool
    file_format: str
    language: str

    @validator("pdf")
    def pdf_path_must_exist(cls, v):
        if not Path(v).exists():
            raise ValueError("pdf_path must exist")
        return v


class Reader(AsyncBaseReader):
    def __init__(
        self,
        key_word,
        query,
        filter_keys,
        root_path=".",
        sort=arxiv.SortCriterion.SubmittedDate,
        user_name="defualt",
        args=None,
    ):
        if args is None:
            raise ValueError("args is None")

        if args.language == "en":
            language = "English"
        elif args.language == "zh":
            language = "Chinese"
        else:
            language = "English"

        super().__init__(
            root_path,
            language,
            args.file_format,
            args.save_image,
        )

        self.user_name = user_name  # 读者姓名
        self.key_word = key_word  # 读者感兴趣的关键词
        self.filter_keys = filter_keys  # 用于在摘要中筛选的关键词
        self.query = query  # 读者输入的搜索查询
        self.sort = sort  # 读者选择的排序方式

    def get_arxiv(self, max_results=30):
        search = arxiv.Search(
            query=self.query,
            max_results=max_results,
            sort_by=self.sort,
            sort_order=arxiv.SortOrder.Descending,
        )
        return search

    def filter_arxiv(self, max_results=30):
        search = self.get_arxiv(max_results=max_results)
        results = list(search.results())

        logger.info("All search:")
        for index, result in enumerate(results):
            logger.info(
                f"{index=}, title={result.title} {result.updated.strftime('%Y-%m-%d')}"
            )

        # if self.filter_keys is empty then do not filter out
        if not self.filter_keys:
            return results

        filter_results = []

        logger.info(f"filter_keys {self.filter_keys}")
        # 确保每个关键词都能在摘要中找到，才算是目标论文
        for index, result in enumerate(search.results()):
            abs_text = result.summary.replace("-\n", "-").replace("\n", " ")
            meet_num = 0
            for f_key in self.filter_keys:
                if f_key.lower() in abs_text.lower():
                    meet_num += 1
            if meet_num == len(self.filter_keys):
                filter_results.append(result)

        logger.info(f"filter_results: {len(filter_results)}")
        logger.info("filter_papers:")
        for index, result in enumerate(filter_results):
            logger.info(
                f"{index=}, title={result.title} {result.updated.strftime('%Y-%m-%d')}"
            )

        return filter_results

    def download_pdf(self, filter_results):
        return asyncio.run(self._download_pdf(filter_results))

    @staticmethod
    def create_paper(results_mapping, paper_path):
        result = results_mapping[paper_path.name]
        paper = Paper(
            path=paper_path,
            url=result.entry_id,
            title=result.title,
            abs=result.summary.replace("-\n", "-").replace("\n", " "),
            authers=[str(aut) for aut in result.authors],
        )
        return paper

    async def _download_pdf(self, filter_results):
        date_str = str(datetime.datetime.now())[:13].replace(" ", "-")
        query_str = (
            self.query.replace("au:", "")
            .replace("title: ", "")
            .replace("ti: ", "")
            .replace(":", " ")[:25]
        )

        path = self.root_path / "pdf_files" / f"{query_str}-{date_str}"

        path.mkdir(parents=True, exist_ok=True)
        logger.info(f"All_paper: {len(filter_results)}")

        paper_list = []
        tasks = []
        results_mapping = {}

        async with aiohttp.ClientSession() as session:
            for _, result in enumerate(filter_results):
                title_str = self.validateTitle(result.title)
                pdf_name = title_str + ".pdf"
                tasks.append(result.download_pdf(session, path.as_posix(), pdf_name))
                results_mapping[pdf_name] = result

            for done_task in asyncio.as_completed(tasks):
                try:
                    paper_path = await done_task
                    paper_list.append(self.create_paper(results_mapping, paper_path))

                except Exception as e:
                    logger.warning(f"download_error: {e}")

        return paper_list

    def show_info(self):
        logger.info(f"Key word: {self.key_word}")
        logger.info(f"Query: {self.query}")
        logger.info(f"Sort: {self.sort}")


def add_subcommand(parser):
    name = "paper"
    subparser = parser.add_parser(
        name, help="Fetch or Summary paper from local or arxiv"
    )

    subparser.add_argument(
        "-p",
        "--pdf",
        type=str,
        default="",
        metavar="",
        help="if none, the bot will download from arxiv with query",
    )

    subparser.add_argument(
        "-q",
        "--query",
        type=str,
        default="all: ChatGPT robot",
        metavar="",
        help="the query string, ti: xx, au: xx, all: xx (default: %(default)s)",
    )

    subparser.add_argument(
        "-k",
        "--key-word",
        type=str,
        default="reinforcement learning",
        metavar="",
        help="the key word of user research fields (default: %(default)s)",
    )

    subparser.add_argument(
        "-m",
        "--max-results",
        type=int,
        default=1,
        metavar="",
        help="the maximum number of results (default: %(default)s)",
    )

    subparser.add_argument(
        "-s",
        "--sort",
        type=str,
        default="Relevance",
        metavar="",
        help="another is LastUpdatedDate (default: %(default)s)",
    )

    subparser.add_argument(
        "-f",
        "--file-format",
        type=str,
        default="md",
        choices=["md", "txt", "pdf", "tex"],
        metavar="",
        help="the format of the exported file, if you save the picture, it is best to be md, if not, the txt will not be messy (default: %(default)s)",
    )

    subparser.add_argument(
        "-l",
        "--language",
        type=str,
        default="en",
        metavar="",
        help="The other output lauguage is English, is en (default: %(default)s)",
    )

    subparser.add_argument(
        "--filter-keys",
        type=str,
        action="extend",
        nargs="+",
        metavar="",
        help="the filter key words, every word in the abstract must have, otherwise it will not be selected as the target paper",
    )

    subparser.add_argument(
        "--save-image",
        default=False,
        metavar="",
        help="save image? It takes a minute or two to save a picture! But pretty (default: %(default)s)",
    )

    return name


def main(args):
    if args.sort == "Relevance":
        sort = arxiv.SortCriterion.Relevance
    elif args.sort == "LastUpdatedDate":
        sort = arxiv.SortCriterion.LastUpdatedDate
    else:
        sort = arxiv.SortCriterion.Relevance

    if args.pdf:
        reader = Reader(
            key_word=args.key_word,
            query=args.query,
            filter_keys=args.filter_keys,
            sort=sort,
            args=args,
        )
        reader.show_info()
        paper_list = []
        if args.pdf.endswith(".pdf"):
            paper_list.append(Paper(path=args.pdf))
            logger.info(f"read pdf file {args.pdf}")
        else:
            logger.info(f"read pdf files from path {args.pdf}")
            for root, dirs, files in os.walk(args.pdf):
                logger.trace(f"root: {root}, dirs: {dirs}, files: {files}")
                for filename in files:
                    if filename.endswith(".pdf"):
                        paper_list.append(Paper(path=os.path.join(root, filename)))

        logger.info("paper_num: {}".format(len(paper_list)))
        for paper_index, paper in enumerate(paper_list):
            name = Path(paper.path).name
            logger.info(f"{paper_index=}, {name=}")
        reader.summary_with_chat(paper_list=paper_list, key_words=reader.key_word)
    else:
        reader = Reader(
            key_word=args.key_word,
            query=args.query,
            filter_keys=args.filter_keys,
            sort=sort,
            args=args,
        )
        reader.show_info()
        filter_results = reader.filter_arxiv(max_results=args.max_results)
        paper_list = reader.download_pdf(filter_results)
        reader.summary_with_chat(paper_list=paper_list, key_words=reader.key_word)

    reader.show_token_usage()


def cli(args):
    parameters = PaperParams(**vars(args))
    main(parameters)



import io
import os

import fitz
from loguru import logger
from PIL import Image


class Paper:
    def __init__(self, path, title="", url="", abs="", authers=[]):
        self.url = url  # 文章链接
        self.path = path  # pdf路径
        self.section_names = []  # 段落标题
        self.section_texts = {}  # 段落内容
        self.abs = abs
        self.title_page = 0
        self.pdf = fitz.open(self.path)  # pdf文档

        self.title = self.get_title() if title == "" else title

        self.parse_pdf()
        self.authers = authers
        self.roman_num = [
            "I",
            "II",
            "III",
            "IV",
            "V",
            "VI",
            "VII",
            "VIII",
            "IIX",
            "IX",
            "X",
        ]

        self.digit_num = [str(d + 1) for d in range(10)]
        self.first_image = ""

    def __repr__(self):
        return f"""Papers(title={self.title}, url={self.url}, authers={self.authers} abs={self.abs})"""

    __str__ = __repr__

    def parse_pdf(self):
        self.pdf = fitz.open(self.path)  # pdf文档
        self.text_list = [page.get_text() for page in self.pdf]
        self.all_text = " ".join(self.text_list)
        self.section_page_dict = self._get_all_page_index()  # 段落与页码的对应字典
        logger.trace(f"section_page_dict {self.section_page_dict}")
        self.section_text_dict = self._get_all_page()  # 段落与内容的对应字典
        self.section_text_dict.update({"title": self.title})
        self.section_text_dict.update({"paper_info": self.get_paper_info()})
        self.pdf.close()

    def get_paper_info(self):
        first_page_text = self.pdf[self.title_page].get_text()
        if "Abstract" in self.section_text_dict.keys():
            abstract_text = self.section_text_dict["Abstract"]
        else:
            abstract_text = self.abs
        first_page_text = first_page_text.replace(abstract_text, "")
        return first_page_text

    def get_image_path(self, image_path=""):
        """
        将PDF中的第一张图保存到image.png里面，存到本地目录，返回文件名称，供gitee读取
        :param filename: 图片所在路径，"C:\\Users\\Administrator\\Desktop\\nwd.pdf"
        :param image_path: 图片提取后的保存路径
        :return:
        """
        max_size = 0
        image_list = []
        ext = None
        with fitz.Document(self.path) as my_pdf_file:
            for page_number in range(1, len(my_pdf_file) + 1):
                page = my_pdf_file[page_number - 1]
                page.get_images()
                for _, image in enumerate(page.get_images(), start=1):
                    xref_value = image[0]
                    base_image = my_pdf_file.extract_image(xref_value)
                    image_bytes = base_image["image"]
                    ext = base_image["ext"]
                    image = Image.open(io.BytesIO(image_bytes))
                    image_size = image.size[0] * image.size[1]
                    if image_size > max_size:
                        max_size = image_size
                    image_list.append(image)

        if ext is None:
            raise Exception("No image found in pdf")

        for image in image_list:
            image_size = image.size[0] * image.size[1]
            if image_size == max_size:
                image_name = f"image.{ext}"
                im_path = os.path.join(image_path, image_name)
                logger.trace(f"image_path: {im_path}")

                max_pix = 480
                min(image.size[0], image.size[1])

                if image.size[0] > image.size[1]:
                    min_pix = int(image.size[1] * (max_pix / image.size[0]))
                    newsize = (max_pix, min_pix)
                else:
                    min_pix = int(image.size[0] * (max_pix / image.size[1]))
                    newsize = (min_pix, max_pix)
                image = image.resize(newsize)

                image.save(open(im_path, "wb"))
                return im_path, ext
        return None, None

    def get_chapter_names(
        self,
    ):
        doc = fitz.open(self.path)  # pdf文档
        text_list = [page.get_text() for page in doc]
        all_text = ""
        for text in text_list:
            all_text += text
        # # 创建一个空列表，用于存储章节名称
        chapter_names = []
        for line in all_text.split("\n"):
            line.split(" ")
            if "." in line:
                point_split_list = line.split(".")
                space_split_list = line.split(" ")
                if 1 < len(space_split_list) < 5:
                    if 1 < len(point_split_list) < 5 and (
                        point_split_list[0] in self.roman_num
                        or point_split_list[0] in self.digit_num
                    ):
                        logger.info(f"{line=}")
                        chapter_names.append(line)
                    # 这段代码可能会有新的bug，本意是为了消除"Introduction"的问题的！
                    elif 1 < len(point_split_list) < 5:
                        logger.info(f"{line=}")
                        chapter_names.append(line)

        return chapter_names

    def fetch_title(self):
        logger.trace(self.pdf.metadata)
        return self.pdf.metadata["title"]

    def get_title(self):
        max_font_size = 0  # 初始化最大字体大小为0
        max_font_sizes = [0]
        for page_index, page in enumerate(self.pdf):  # 遍历每一页
            text = page.get_text("dict")  # 获取页面上的文本信息
            blocks = text["blocks"]  # 获取文本块列表
            for block in blocks:  # 遍历每个文本块
                if block["type"] == 0 and len(block["lines"]):  # 如果是文字类型
                    if len(block["lines"][0]["spans"]):
                        font_size = block["lines"][0]["spans"][0][
                            "size"
                        ]  # 获取第一行第一段文字的字体大小
                        max_font_sizes.append(font_size)
                        if font_size > max_font_size:  # 如果字体大小大于当前最大值
                            max_font_size = font_size  # 更新最大值
                            block["lines"][0]["spans"][0]["text"]  # 更新最大值对应的字符串
        max_font_sizes.sort()
        logger.trace(f"max_font_sizes {max_font_sizes[-10:]}")
        cur_title = ""

        for page_index, page in enumerate(self.pdf):  # 遍历每一页
            text = page.get_text("dict")  # 获取页面上的文本信息
            blocks = text["blocks"]  # 获取文本块列表
            for block in blocks:  # 遍历每个文本块
                if block["type"] == 0 and len(block["lines"]):  # 如果是文字类型
                    if len(block["lines"][0]["spans"]):
                        cur_string = block["lines"][0]["spans"][0][
                            "text"
                        ]  # 更新最大值对应的字符串
                        block["lines"][0]["spans"][0]["flags"]  # 获取第一行第一段文字的字体特征
                        font_size = block["lines"][0]["spans"][0][
                            "size"
                        ]  # 获取第一行第一段文字的字体大小
                        # print(font_size)
                        if (
                            abs(font_size - max_font_sizes[-1]) < 0.3
                            or abs(font_size - max_font_sizes[-2]) < 0.3
                        ):
                            # print("The string is bold.", max_string, "font_size:", font_size, "font_flags:", font_flags)
                            if len(cur_string) > 4 and "arXiv" not in cur_string:
                                # print("The string is bold.", max_string, "font_size:", font_size, "font_flags:", font_flags)
                                if cur_title == "":
                                    cur_title += cur_string
                                else:
                                    cur_title += " " + cur_string
                            self.title_page = page_index

        title = cur_title.replace("\n", " ")
        return title

    def _get_all_page_index(self):
        # define the chapter name list
        section_list = [
            "Abstract",
            "Introduction",
            "Related Work",
            "Background",
            "Introduction and Motivation",
            "Computation Function",
            "Routing Function",
            "Preliminary",
            "Problem Formulation",
            "Methods",
            "Methodology",
            "Method",
            "Approach",
            "Approaches",
            "Materials and Methods",
            "Experiment Settings",
            "Experiment",
            "Experimental Results",
            "Evaluation",
            "Experiments",
            "Results",
            "Findings",
            "Data Analysis",
            "Discussion",
            "Results and Discussion",
            "Conclusion",
            "References",
        ]
        section_page_dict = {}
        for page_index, page in enumerate(self.pdf):
            cur_text = page.get_text()
            for section_name in section_list:
                section_name_upper = section_name.upper()
                if "Abstract" == section_name and section_name in cur_text:
                    section_page_dict[section_name] = page_index
                else:
                    if section_name + "\n" in cur_text:
                        section_page_dict[section_name] = page_index
                    elif section_name_upper + "\n" in cur_text:
                        section_page_dict[section_name] = page_index
        return section_page_dict

    def _get_all_page(self):
        """

        Get the text information of each page in the PDF file and return the text information as a dictionary according to the chapter.

        Returns:
            section_dict (dict): Text information dictionary for each chapter, key for chapter name, value for chapter text.

        """
        text_list = []
        section_dict = {}

        text_list = [page.get_text() for page in self.pdf]
        for sec_index, sec_name in enumerate(self.section_page_dict):
            logger.trace(
                f"{sec_index=}, {sec_name=}, {self.section_page_dict[sec_name]}"
            )
            if sec_index <= 0 and self.abs:
                continue
            else:
                start_page = self.section_page_dict[sec_name]
                if sec_index < len(list(self.section_page_dict.keys())) - 1:
                    end_page = self.section_page_dict[
                        list(self.section_page_dict.keys())[sec_index + 1]
                    ]
                else:
                    end_page = len(text_list)

                logger.trace(f"{start_page=}, {end_page=}")

                cur_sec_text = ""
                if end_page - start_page == 0:
                    if sec_index < len(list(self.section_page_dict.keys())) - 1:
                        next_sec = list(self.section_page_dict.keys())[sec_index + 1]
                        if text_list[start_page].find(sec_name) == -1:
                            start_i = text_list[start_page].find(sec_name.upper())
                        else:
                            start_i = text_list[start_page].find(sec_name)
                        if text_list[start_page].find(next_sec) == -1:
                            end_i = text_list[start_page].find(next_sec.upper())
                        else:
                            end_i = text_list[start_page].find(next_sec)
                        cur_sec_text += text_list[start_page][start_i:end_i]
                else:
                    for page_i in range(start_page, end_page):
                        #                         print("page_i:", page_i)
                        if page_i == start_page:
                            if text_list[start_page].find(sec_name) == -1:
                                start_i = text_list[start_page].find(sec_name.upper())
                            else:
                                start_i = text_list[start_page].find(sec_name)
                            cur_sec_text += text_list[page_i][start_i:]
                        elif page_i < end_page:
                            cur_sec_text += text_list[page_i]
                        elif page_i == end_page:
                            if sec_index < len(list(self.section_page_dict.keys())) - 1:
                                next_sec = list(self.section_page_dict.keys())[
                                    sec_index + 1
                                ]
                                if text_list[start_page].find(next_sec) == -1:
                                    end_i = text_list[start_page].find(next_sec.upper())
                                else:
                                    end_i = text_list[start_page].find(next_sec)
                                cur_sec_text += text_list[page_i][:end_i]
                section_dict[sec_name] = cur_sec_text.replace("-\n", "").replace(
                    "\n", " "
                )
        return section_dict



import asyncio
import json
import re
import time
import typing as t
from calendar import timegm
from datetime import datetime, timezone
from enum import Enum, auto
from pathlib import Path
from typing import Any, Optional

import aiofiles
import aiohttp
import feedparser
import requests
import tenacity
from loguru import logger
from pydantic import BaseModel

_DEFAULT_TIME = datetime.min


class Category(Enum):
    AnimalBeHaviorAndCognition = auto()
    Biochemistry = auto()
    Bioengineering = auto()
    Bioinformatics = auto()
    Biophysics = auto()
    CancerBiology = auto()
    CellBiology = auto()
    ClinicalTrials = auto()
    DevelopmentalBiology = auto()
    Ecology = auto()
    Epidemiology = auto()
    EvolutionaryBiology = auto()
    Genetics = auto()
    Genomics = auto()
    Immunology = auto()
    Microbiology = auto()
    MolecularBiology = auto()
    Neuroscience = auto()
    Paleontology = auto()
    Pathology = auto()
    PharmacologyAndToxicology = auto()
    Physiology = auto()
    PlantBiology = auto()
    ScientificCommunicationAndEducation = auto()
    SyntheticBiology = auto()
    SystemsBiology = auto()
    Zoology = auto()
    Unknown = auto()

    @classmethod
    def from_str(cls, category) -> "Category":
        if category == "animal behavior and cogtition":
            return cls.AnimalBeHaviorAndCognition
        elif category == "biochemistry":
            return cls.Biochemistry
        elif category == "bioengineering":
            return cls.Bioengineering
        elif category == "bioinformatics":
            return cls.Bioinformatics
        elif category == "biophysics":
            return cls.Biophysics
        elif category == "cancer biology":
            return cls.CancerBiology
        elif category == "clinical trials":
            return cls.ClinicalTrials
        elif category == "developmental biology":
            return cls.DevelopmentalBiology
        elif category == "ecology":
            return cls.Ecology
        elif category == "epidemiology":
            return cls.Epidemiology
        elif category == "evolutionary biology":
            return cls.EvolutionaryBiology
        elif category == "genetics":
            return cls.Genetics
        elif category == "genomics":
            return cls.Genomics
        elif category == "immunology":
            return cls.Immunology
        elif category == "microbiology":
            return cls.Microbiology
        elif category == "molecular biology":
            return cls.MolecularBiology
        elif category == "neuroscience":
            return cls.Neuroscience
        elif category == "paleontology":
            return cls.Paleontology
        elif category == "pathology":
            return cls.Pathology
        elif category == "pharmacology and toxicology":
            return cls.PharmacologyAndToxicology
        elif category == "physiology":
            return cls.Physiology
        elif category == "plant biology":
            return cls.PlantBiology
        elif category == "scientific communication and education":
            return cls.ScientificCommunicationAndEducation
        elif category == "synthetic biology":
            return cls.SyntheticBiology
        elif category == "systems biology":
            return cls.SystemsBiology
        elif category == "zoology":
            return cls.Zoology
        else:
            return cls.Unknown


class Result:
    """
        An entry in an bioarXiv query results feed.

    Attributes
        start_date: str = None
        end_date: str = None
        days: int = None
        max_results: float
        sort_by: SortCriterion = SortCriterion.Relevance
        sort_order: SortOrder = SortOrder.Descending
    ```

    """

    def __init__(
        self,
        doi: str,
        title: str,
        authors: str,
        author_corresponding: str,
        author_corresponding_institution: str,
        version: int,
        category: Category,
        jats_xml_path: str,
        abstract: str,
        published: str,
        server: str,
        pdf_url: Optional[str] = None,
        entry_id: Optional[str] = None,
        date: datetime = _DEFAULT_TIME,
    ):
        self.doi = Result.validate_doi(doi)
        self.title = title
        self.authors = authors
        self.author_corresponding = author_corresponding
        self.author_corresponding_institution = author_corresponding_institution
        self.version = version
        self.category = category
        self.jats_xml_path = Result.validate_jats_xml(jats_xml_path)
        self.abstract = abstract
        self.published = published
        self.server = server
        self.pdf_url = pdf_url
        self.entry_id = entry_id
        self.date = Result.validate_date(date)

        self.headers = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.163 "
            "Safari/537.36",
            "Connection": "keep-alive",
        }

    @staticmethod
    def validate_doi(doi):
        """Validate a DOI."""
        return doi.replace("\\", "")

    @staticmethod
    def validate_jats_xml(jats_xml):
        """Validate a JATS XML URL."""
        return jats_xml.replace("\\", "")

    @staticmethod
    def validate_date(date):
        """Validate a date."""
        return Result._to_datetime(time.strptime(date, "%Y-%m-%d"))

    @classmethod
    def from_api_entry(cls, entry) -> "Result":
        try:
            result = cls(
                doi=entry["doi"],
                title=entry["title"],
                authors=entry["authors"],
                author_corresponding=entry["author_corresponding"],
                author_corresponding_institution=entry[
                    "author_corresponding_institution"
                ],
                date=entry["date"],
                version=entry["version"],
                category=Category.from_str(entry["category"]),
                jats_xml_path=entry["jatsxml"],
                abstract=entry["abstract"],
                published=entry["published"],
                server=entry["server"],
            )
        except KeyError as e:
            raise MissingFieldError(f"{e}")
        else:
            return result

    def __repr__(self) -> str:
        return """Result(doi={}, title={}, authors={}, author_corresponding={},
                         author_corresponding_institution={}, date={}, version={},
                         category={}, jats_xml_path={}, abstract={}, published={}, server={})""".format(
            self.doi,
            self.title,
            self.authors,
            self.author_corresponding,
            self.author_corresponding_institution,
            self.date,
            self.version,
            self.category,
            self.jats_xml_path,
            self.abstract,
            self.published,
            self.server,
        )

    def __eq__(self, other) -> bool:
        if isinstance(other, Result):
            return self.doi == other.doi
        return False

    def get_short_id(self) -> str:
        """
        Returns the short ID for this result.

        + If the result URL is `"http://arxiv.org/abs/2107.05580v1"`,
        `result.get_short_id()` returns `2107.05580v1`.

        + If the result URL is `"http://arxiv.org/abs/quant-ph/0201082v1"`,
        `result.get_short_id()` returns `"quant-ph/0201082v1"` (the pre-March
        2007 arXiv identifier format).

        For an explanation of the difference between arXiv's legacy and current
        identifiers, see [Understanding the arXiv
        identifier](https://arxiv.org/help/arxiv_identifier).
        """
        return self.doi

    async def get_pdf_url(self, session: aiohttp.ClientSession) -> str:
        doi_base = "https://doi.org"
        self.entry_id = f"{doi_base}/{self.doi}"

        async with session.get(self.entry_id, headers=self.headers) as response:
            content_path = f"{response.url}.pdf"
            await asyncio.sleep(0.001)
            return content_path

    def _get_default_filename(self, extension: str = "pdf") -> str:
        """
        A default `to_filename` function for the extension given.
        """
        nonempty_title = self.title if self.title else "UNTITLED"
        # Remove disallowed characters.
        clean_title = "_".join(re.findall(r"\w+", nonempty_title))
        return "{}.{}.{}".format(self.get_short_id(), clean_title, extension)

    async def download_pdf(
        self,
        session: aiohttp.ClientSession,
        dirpath: Path = Path("."),
        filename: str = "",
    ) -> Path:
        """
        Downloads the PDF for this result to the specified directory.

        The filename is generated by calling `to_filename(self)`.
        """
        if isinstance(dirpath, str):
            dirpath = Path(dirpath)

        if not filename:
            filename = self._get_default_filename()

        path = dirpath / filename

        self.pdf_url = await self.get_pdf_url(session)
        logger.info(f"Downloading: {filename}")
        async with session.get(
            self.pdf_url, headers=self.headers, raise_for_status=True
        ) as response:
            async with aiofiles.open(path, "wb") as f:
                async for chunk in response.content.iter_chunked(1024):
                    await asyncio.sleep(0.001)
                    await f.write(chunk)

        return path

    @staticmethod
    def _to_datetime(ts: time.struct_time) -> datetime:
        """
        Converts a UTC time.struct_time into a time-zone-aware datetime.

        This will be replaced with feedparser functionality [when it becomes
        available](https://github.com/kurtmckee/feedparser/issues/212).
        """
        return datetime.fromtimestamp(timegm(ts), tz=timezone.utc)


class MissingFieldError(Exception):
    """
    An error indicating an entry is unparseable because it lacks required
    fields.
    """

    missing_field: str
    """The required field missing from the would-be entry."""
    message: str
    """Message describing what caused this error."""

    def __init__(self, missing_field):
        self.missing_field = missing_field
        self.message = "Entry from BiorXiv missing required info"

    def __repr__(self) -> str:
        return "{}({})".format(_classname(self), repr(self.missing_field))


class SortCriterion(Enum):
    """
    A SortCriterion identifies a property by which search results can be
    sorted.

    See [the arXiv API User's Manual: sort order for return
    results](https://arxiv.org/help/api/user-manual#sort).
    """

    Relevance = "relevance"
    LastUpdatedDate = "lastUpdatedDate"
    SubmittedDate = "submittedDate"


class SortOrder(Enum):
    """
    A SortOrder indicates order in which search results are sorted according
    to the specified arxiv.SortCriterion.

    See [the arXiv API User's Manual: sort order for return
    results](https://arxiv.org/help/api/user-manual#sort).
    """

    Ascending = "ascending"
    Descending = "descending"


class Search(BaseModel):
    """
    A specification for a search of arXiv's database.

    To run a search, use `Search.run` to use a default client or `Client.run`
    with a specific client.
    """

    start_date: Optional[str] = None
    end_date: Optional[str] = None
    days: Optional[int] = None
    server: str
    max_results: float
    """
    The maximum number of results to be returned in an execution of this
    search.

    To fetch every result available, set `max_results=float('inf')`.
    """

    sort_by: SortCriterion = SortCriterion.Relevance
    """The sort criterion for results."""
    sort_order: SortOrder = SortOrder.Descending
    """The sort order for results."""

    def __repr__(self) -> str:
        return "Search(start-date={}, end-date={}, days={}, max_results={})".format(
            self.start_date,
            self.end_date,
            self.days,
            self.max_results,
        )

    # async def _result(
    #     self,
    #     offset: int,
    # ):
    #     async with aiohttp.ClientSession() as session:
    #         results = list(Client().results(self, offset))
    #         tasks = []
    #         for result in results:
    #             tasks.append(result.download_pdf(session))

    #         await asyncio.gather(*tasks)
    #         return results

    # def results(self, offset: int = 0):
    #     """
    #     Executes the specified search using a default arXiv API client.

    #     For info on default behavior, see `Client.__init__` and `Client.results`.
    #     """

    #     return asyncio.run(self._result(offset))

    def results(self, offset: int = 0) -> t.Generator[Result, None, None]:
        """
        Executes the specified search using a default arXiv API client.

        For info on default behavior, see `Client.__init__` and `Client.results`.
        """
        return Client().results(self, offset=offset)


class Client(BaseModel):
    """
    Specifies a strategy for fetching results from arXiv's API.

    This class obscures pagination and retry logic, and exposes
    `Client.results`.

    query_url_format = "http://export.arxiv.org/api/query?{}"
    The arXiv query API endpoint format.
    page_size: int
    Maximum number of results fetched in a single API request.
    delay_seconds: int
    Number of seconds to wait between API requests.
    num_retries: int
    Number of times to retry a failing API request.
    _last_request_dt: datetime
    """

    biorxiv_api: str = "https://api.biorxiv.org/details/biorxiv"
    delay_seconds: int = 3
    num_retries: int = 3
    _last_request_dt: Optional[datetime] = None

    def __repr__(self) -> str:
        return "Client(delay_seconds={}, num_retries={})".format(
            self.delay_seconds,
            self.num_retries,
        )

    def results(self, search: Search, offset: int = 0):
        """
        Uses this client configuration to fetch one page of the search results
        at a time, yielding the parsed `Result`s, until `max_results` results
        have been yielded or there are no more search results.

        If all tries fail, raises an `UnexpectedEmptyPageError` or `HTTPError`.

        Setting a nonzero `offset` discards leading records in the result set.
        When `offset` is greater than or equal to `search.max_results`, the full
        result set is discarded.

        For more on using generators, see
        [Generators](https://wiki.python.org/moin/Generators).
        """

        # total_results may be reduced according to the feed's
        total_results = search.max_results
        entries = self._request(self._format_url(search))

        for entry in entries["collection"]:
            if offset < total_results:
                entry_result = Result.from_api_entry(entry)
                total_results -= 1
                yield entry_result
            else:
                break

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    def _request(self, url) -> dict[str, Any]:
        response = requests.get(url)
        response.raise_for_status()
        content = response.content.strip().decode()
        content = json.loads(content)
        return content

    def _format_url(self, search: Search) -> str:
        """
        Construct a request API for search that returns up to `page_size`
        results starting with the result at index `start`.
        """

        return (
            f"{self.biorxiv_api}/{search.days}d"
            if search.days is not None
            else f"{self.biorxiv_api}/{search.start_date}/{search.end_date}"
        )


class ArxivError(Exception):
    """This package's base Exception class."""

    url: str
    """The feed URL that could not be fetched."""
    retry: int
    """
    The request try number which encountered this error; 0 for the initial try,
    1 for the first retry, and so on.
    """
    message: str
    """Message describing what caused this error."""

    def __init__(self, url: str, retry: int, message: str):
        """
        Constructs an `ArxivError` encountered while fetching the specified URL.
        """
        self.url = url
        self.retry = retry
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return "{} ({})".format(self.message, self.url)


class UnexpectedEmptyPageError(ArxivError):
    """
    An error raised when a page of results that should be non-empty is empty.

    This should never happen in theory, but happens sporadically due to
    brittleness in the underlying arXiv API; usually resolved by retries.

    See `Client.results` for usage.
    """

    def __init__(self, url: str, retry: int):
        """
        Constructs an `UnexpectedEmptyPageError` encountered for the specified
        API URL after `retry` tries.
        """
        self.url = url
        super().__init__(url, retry, "Page of results was unexpectedly empty")

    def __repr__(self) -> str:
        return "{}({}, {})".format(_classname(self), repr(self.url), repr(self.retry))


class HTTPError(ArxivError):
    """
    A non-200 status encountered while fetching a page of results.

    See `Client.results` for usage.
    """

    status: int
    """The HTTP status reported by feedparser."""
    entry: feedparser.FeedParserDict
    """The feed entry describing the error, if present."""

    def __init__(self, url: str, retry: int, feed: feedparser.FeedParserDict):
        """
        Constructs an `HTTPError` for the specified status code, encountered for
        the specified API URL after `retry` tries.
        """
        self.url = url
        self.status = feed.status
        # If the feed is valid and includes a single entry, trust it's an
        # explanation.
        if not feed.bozo and len(feed.entries) == 1:
            self.entry = feed.entries[0]
        else:
            self.entry = None
        super().__init__(
            url,
            retry,
            "Page request resulted in HTTP {}: {}".format(
                self.status,
                self.entry.summary if self.entry else None,
            ),
        )

    def __repr__(self) -> str:
        return "{}({}, {}, {})".format(
            _classname(self), repr(self.url), repr(self.retry), repr(self.status)
        )


def _classname(o):
    """A helper function for use in __repr__ methods: arxiv.Result.Link."""
    return "{}".format(o.__class__.__qualname__)



import aiofiles

from ..utils import GLOBAL_PATH

EISVOGEL = r"""
%%
% Copyright (c) 2017 - 2023, Pascal Wagler;
% Copyright (c) 2014 - 2023, John MacFarlane
%
% All rights reserved.
%
% Redistribution and use in source and binary forms, with or without
% modification, are permitted provided that the following conditions
% are met:
%
% - Redistributions of source code must retain the above copyright
% notice, this list of conditions and the following disclaimer.
%
% - Redistributions in binary form must reproduce the above copyright
% notice, this list of conditions and the following disclaimer in the
% documentation and/or other materials provided with the distribution.
%
% - Neither the name of John MacFarlane nor the names of other
% contributors may be used to endorse or promote products derived
% from this software without specific prior written permission.
%
% THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
% "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
% LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
% FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
% COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
% INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
% BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
% LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
% CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
% LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
% ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
% POSSIBILITY OF SUCH DAMAGE.
%%

%%
% This is the Eisvogel pandoc LaTeX template.
%
% For usage information and examples visit the official GitHub page:
% https://github.com/Wandmalfarbe/pandoc-latex-template
%%

% Options for packages loaded elsewhere
\PassOptionsToPackage{unicode$for(hyperrefoptions)$,$hyperrefoptions$$endfor$}{hyperref}
\PassOptionsToPackage{hyphens}{url}
\PassOptionsToPackage{dvipsnames,svgnames,x11names,table}{xcolor}
$if(CJKmainfont)$
\PassOptionsToPackage{space}{xeCJK}
$endif$
%
\documentclass[
$if(fontsize)$
  $fontsize$,
$endif$
$if(papersize)$
  $papersize$paper,
$else$
  paper=a4,
$endif$
$if(beamer)$
  ignorenonframetext,
$if(handout)$
  handout,
$endif$
$if(aspectratio)$
  aspectratio=$aspectratio$,
$endif$
$endif$
$for(classoption)$
  $classoption$$sep$,
$endfor$
  ,captions=tableheading
]{$if(beamer)$$documentclass$$else$$if(book)$scrbook$else$scrartcl$endif$$endif$}
$if(beamer)$
$if(background-image)$
\usebackgroundtemplate{%
  \includegraphics[width=\paperwidth]{$background-image$}%
}
% In beamer background-image does not work well when other images are used, so this is the workaround
\pgfdeclareimage[width=\paperwidth,height=\paperheight]{background}{$background-image$}
\usebackgroundtemplate{\pgfuseimage{background}}
$endif$
\usepackage{pgfpages}
\setbeamertemplate{caption}[numbered]
\setbeamertemplate{caption label separator}{: }
\setbeamercolor{caption name}{fg=normal text.fg}
\beamertemplatenavigationsymbols$if(navigation)$$navigation$$else$empty$endif$
$for(beameroption)$
\setbeameroption{$beameroption$}
$endfor$
% Prevent slide breaks in the middle of a paragraph
\widowpenalties 1 10000
\raggedbottom
$if(section-titles)$
\setbeamertemplate{part page}{
  \centering
  \begin{beamercolorbox}[sep=16pt,center]{part title}
    \usebeamerfont{part title}\insertpart\par
  \end{beamercolorbox}
}
\setbeamertemplate{section page}{
  \centering
  \begin{beamercolorbox}[sep=12pt,center]{part title}
    \usebeamerfont{section title}\insertsection\par
  \end{beamercolorbox}
}
\setbeamertemplate{subsection page}{
  \centering
  \begin{beamercolorbox}[sep=8pt,center]{part title}
    \usebeamerfont{subsection title}\insertsubsection\par
  \end{beamercolorbox}
}
\AtBeginPart{
  \frame{\partpage}
}
\AtBeginSection{
  \ifbibliography
  \else
    \frame{\sectionpage}
  \fi
}
\AtBeginSubsection{
  \frame{\subsectionpage}
}
$endif$
$endif$
$if(beamerarticle)$
\usepackage{beamerarticle} % needs to be loaded first
$endif$
\usepackage{amsmath,amssymb}
$if(linestretch)$
\usepackage{setspace}
$else$
% Use setspace anyway because we change the default line spacing.
% The spacing is changed early to affect the titlepage and the TOC.
\usepackage{setspace}
\setstretch{1.2}
$endif$
\usepackage{iftex}
\ifPDFTeX
  \usepackage[$if(fontenc)$$fontenc$$else$T1$endif$]{fontenc}
  \usepackage[utf8]{inputenc}
  \usepackage{textcomp} % provide euro and other symbols
\else % if luatex or xetex
$if(mathspec)$
  \ifXeTeX
    \usepackage{mathspec} % this also loads fontspec
  \else
    \usepackage{unicode-math} % this also loads fontspec
  \fi
$else$
  \usepackage{unicode-math} % this also loads fontspec
$endif$
  \defaultfontfeatures{Scale=MatchLowercase}$-- must come before Beamer theme
  \defaultfontfeatures[\rmfamily]{Ligatures=TeX,Scale=1}
\fi
$if(fontfamily)$
$else$
$-- Set default font before Beamer theme so the theme can override it
\usepackage{lmodern}
$endif$
$-- Set Beamer theme before user font settings so they can override theme
$if(beamer)$
$if(theme)$
\usetheme[$for(themeoptions)$$themeoptions$$sep$,$endfor$]{$theme$}
$endif$
$if(colortheme)$
\usecolortheme{$colortheme$}
$endif$
$if(fonttheme)$
\usefonttheme{$fonttheme$}
$endif$
$if(mainfont)$
\usefonttheme{serif} % use mainfont rather than sansfont for slide text
$endif$
$if(innertheme)$
\useinnertheme{$innertheme$}
$endif$
$if(outertheme)$
\useoutertheme{$outertheme$}
$endif$
$endif$
$-- User font settings (must come after default font and Beamer theme)
$if(fontfamily)$
\usepackage[$for(fontfamilyoptions)$$fontfamilyoptions$$sep$,$endfor$]{$fontfamily$}
$endif$
\ifPDFTeX\else
  % xetex/luatex font selection
$if(mainfont)$
  \setmainfont[$for(mainfontoptions)$$mainfontoptions$$sep$,$endfor$]{$mainfont$}
$endif$
$if(sansfont)$
  \setsansfont[$for(sansfontoptions)$$sansfontoptions$$sep$,$endfor$]{$sansfont$}
$endif$
$if(monofont)$
  \setmonofont[$for(monofontoptions)$$monofontoptions$$sep$,$endfor$]{$monofont$}
$endif$
$for(fontfamilies)$
  \newfontfamily{$fontfamilies.name$}[$for(fontfamilies.options)$$fontfamilies.options$$sep$,$endfor$]{$fontfamilies.font$}
$endfor$
$if(mathfont)$
$if(mathspec)$
  \ifXeTeX
    \setmathfont(Digits,Latin,Greek)[$for(mathfontoptions)$$mathfontoptions$$sep$,$endfor$]{$mathfont$}
  \else
    \setmathfont[$for(mathfontoptions)$$mathfontoptions$$sep$,$endfor$]{$mathfont$}
  \fi
$else$
  \setmathfont[$for(mathfontoptions)$$mathfontoptions$$sep$,$endfor$]{$mathfont$}
$endif$
$endif$
$if(CJKmainfont)$
  \ifXeTeX
    \usepackage{xeCJK}
    \setCJKmainfont[$for(CJKoptions)$$CJKoptions$$sep$,$endfor$]{$CJKmainfont$}
    $if(CJKsansfont)$
      \setCJKsansfont[$for(CJKoptions)$$CJKoptions$$sep$,$endfor$]{$CJKsansfont$}
    $endif$
    $if(CJKmonofont)$
      \setCJKmonofont[$for(CJKoptions)$$CJKoptions$$sep$,$endfor$]{$CJKmonofont$}
    $endif$
  \fi
$endif$
$if(luatexjapresetoptions)$
  \ifLuaTeX
    \usepackage[$for(luatexjapresetoptions)$$luatexjapresetoptions$$sep$,$endfor$]{luatexja-preset}
  \fi
$endif$
$if(CJKmainfont)$
  \ifLuaTeX
    \usepackage[$for(luatexjafontspecoptions)$$luatexjafontspecoptions$$sep$,$endfor$]{luatexja-fontspec}
    \setmainjfont[$for(CJKoptions)$$CJKoptions$$sep$,$endfor$]{$CJKmainfont$}
  \fi
$endif$
\fi
$if(zero-width-non-joiner)$
%% Support for zero-width non-joiner characters.
\makeatletter
\def\zerowidthnonjoiner{%
  % Prevent ligatures and adjust kerning, but still support hyphenating.
  \texorpdfstring{%
    \TextOrMath{\nobreak\discretionary{-}{}{\kern.03em}%
      \ifvmode\else\nobreak\hskip\z@skip\fi}{}%
  }{}%
}
\makeatother
\ifPDFTeX
  \DeclareUnicodeCharacter{200C}{\zerowidthnonjoiner}
\else
  \catcode`^^^^200c=\active
  \protected\def ^^^^200c{\zerowidthnonjoiner}
\fi
%% End of ZWNJ support
$endif$
% Use upquote if available, for straight quotes in verbatim environments
\IfFileExists{upquote.sty}{\usepackage{upquote}}{}
\IfFileExists{microtype.sty}{% use microtype if available
  \usepackage[$for(microtypeoptions)$$microtypeoptions$$sep$,$endfor$]{microtype}
  \UseMicrotypeSet[protrusion]{basicmath} % disable protrusion for tt fonts
}{}
$if(indent)$
$else$
\makeatletter
\@ifundefined{KOMAClassName}{% if non-KOMA class
  \IfFileExists{parskip.sty}{%
    \usepackage{parskip}
  }{% else
    \setlength{\parindent}{0pt}
    \setlength{\parskip}{6pt plus 2pt minus 1pt}}
}{% if KOMA class
  \KOMAoptions{parskip=half}}
\makeatother
$endif$
$if(verbatim-in-note)$
\usepackage{fancyvrb}
$endif$
\usepackage{xcolor}
\definecolor{default-linkcolor}{HTML}{A50000}
\definecolor{default-filecolor}{HTML}{A50000}
\definecolor{default-citecolor}{HTML}{4077C0}
\definecolor{default-urlcolor}{HTML}{4077C0}
$if(footnotes-pretty)$
% load footmisc in order to customize footnotes (footmisc has to be loaded before hyperref, cf. https://tex.stackexchange.com/a/169124/144087)
\usepackage[hang,flushmargin,bottom,multiple]{footmisc}
\setlength{\footnotemargin}{0.8em} % set space between footnote nr and text
\setlength{\footnotesep}{\baselineskip} % set space between multiple footnotes
\setlength{\skip\footins}{0.3cm} % set space between page content and footnote
\setlength{\footskip}{0.9cm} % set space between footnote and page bottom
$endif$
$if(geometry)$
$if(beamer)$
\geometry{$for(geometry)$$geometry$$sep$,$endfor$}
$else$
\usepackage[$for(geometry)$$geometry$$sep$,$endfor$]{geometry}
$endif$
$else$
$if(beamer)$
$else$
\usepackage[margin=2.5cm,includehead=true,includefoot=true,centering,$for(geometry)$$geometry$$sep$,$endfor$]{geometry}
$endif$
$endif$
$if(titlepage-logo)$
\usepackage[export]{adjustbox}
\usepackage{graphicx}
$endif$
$if(beamer)$
\newif\ifbibliography
$endif$
$if(listings)$
\usepackage{listings}
\newcommand{\passthrough}[1]{#1}
\lstset{defaultdialect=[5.3]Lua}
\lstset{defaultdialect=[x86masm]Assembler}
$endif$
$if(listings-no-page-break)$
\usepackage{etoolbox}
\BeforeBeginEnvironment{lstlisting}{\par\noindent\begin{minipage}{\linewidth}}
\AfterEndEnvironment{lstlisting}{\end{minipage}\par\addvspace{\topskip}}
$endif$
$if(lhs)$
\lstnewenvironment{code}{\lstset{language=Haskell,basicstyle=\small\ttfamily}}{}
$endif$
$if(highlighting-macros)$
$highlighting-macros$

% Workaround/bugfix from jannick0.
% See https://github.com/jgm/pandoc/issues/4302#issuecomment-360669013)
% or https://github.com/Wandmalfarbe/pandoc-latex-template/issues/2
%
% Redefine the verbatim environment 'Highlighting' to break long lines (with
% the help of fvextra). Redefinition is necessary because it is unlikely that
% pandoc includes fvextra in the default template.
\usepackage{fvextra}
\DefineVerbatimEnvironment{Highlighting}{Verbatim}{breaklines,fontsize=$if(code-block-font-size)$$code-block-font-size$$else$\small$endif$,commandchars=\\\{\}}

$endif$
$if(tables)$
\usepackage{longtable,booktabs,array}
$if(multirow)$
\usepackage{multirow}
$endif$
\usepackage{calc} % for calculating minipage widths
$if(beamer)$
\usepackage{caption}
% Make caption package work with longtable
\makeatletter
\def\fnum@table{\tablename~\thetable}
\makeatother
$else$
% Correct order of tables after \paragraph or \subparagraph
\usepackage{etoolbox}
\makeatletter
\patchcmd\longtable{\par}{\if@noskipsec\mbox{}\fi\par}{}{}
\makeatother
% Allow footnotes in longtable head/foot
\IfFileExists{footnotehyper.sty}{\usepackage{footnotehyper}}{\usepackage{footnote}}
\makesavenoteenv{longtable}
$endif$
$endif$
% add backlinks to footnote references, cf. https://tex.stackexchange.com/questions/302266/make-footnote-clickable-both-ways
$if(footnotes-disable-backlinks)$
$else$
\usepackage{footnotebackref}
$endif$
$if(graphics)$
\usepackage{graphicx}
\makeatletter
\def\maxwidth{\ifdim\Gin@nat@width>\linewidth\linewidth\else\Gin@nat@width\fi}
\def\maxheight{\ifdim\Gin@nat@height>\textheight\textheight\else\Gin@nat@height\fi}
\makeatother
% Scale images if necessary, so that they will not overflow the page
% margins by default, and it is still possible to overwrite the defaults
% using explicit options in \includegraphics[width, height, ...]{}
\setkeys{Gin}{width=\maxwidth,height=\maxheight,keepaspectratio}
% Set default figure placement to htbp
\makeatletter
\def\fps@figure{htbp}
\makeatother
$endif$
$if(svg)$
\usepackage{svg}
$endif$
$if(strikeout)$
$-- also used for underline
\usepackage{soul}
$endif$
\setlength{\emergencystretch}{3em} % prevent overfull lines
\providecommand{\tightlist}{%
  \setlength{\itemsep}{0pt}\setlength{\parskip}{0pt}}
$if(numbersections)$
\setcounter{secnumdepth}{$if(secnumdepth)$$secnumdepth$$else$5$endif$}
$else$
\setcounter{secnumdepth}{-\maxdimen} % remove section numbering
$endif$
$if(subfigure)$
\usepackage{subcaption}
$endif$
$if(beamer)$
$else$
$if(block-headings)$
% Make \paragraph and \subparagraph free-standing
\ifx\paragraph\undefined\else
  \let\oldparagraph\paragraph
  \renewcommand{\paragraph}[1]{\oldparagraph{#1}\mbox{}}
\fi
\ifx\subparagraph\undefined\else
  \let\oldsubparagraph\subparagraph
  \renewcommand{\subparagraph}[1]{\oldsubparagraph{#1}\mbox{}}
\fi
$endif$
$endif$
$if(pagestyle)$
\pagestyle{$pagestyle$}
$endif$
$if(csl-refs)$
\newlength{\cslhangindent}
\setlength{\cslhangindent}{1.5em}
\newlength{\csllabelwidth}
\setlength{\csllabelwidth}{3em}
\newlength{\cslentryspacingunit} % times entry-spacing
\setlength{\cslentryspacingunit}{\parskip}
\newenvironment{CSLReferences}[2] % #1 hanging-ident, #2 entry spacing
 {% don't indent paragraphs
  \setlength{\parindent}{0pt}
  % turn on hanging indent if param 1 is 1
  \ifodd #1
  \let\oldpar\par
  \def\par{\hangindent=\cslhangindent\oldpar}
  \fi
  % set entry spacing
  \setlength{\parskip}{#2\cslentryspacingunit}
 }%
 {}
\usepackage{calc}
\newcommand{\CSLBlock}[1]{#1\hfill\break}
\newcommand{\CSLLeftMargin}[1]{\parbox[t]{\csllabelwidth}{#1}}
\newcommand{\CSLRightInline}[1]{\parbox[t]{\linewidth - \csllabelwidth}{#1}\break}
\newcommand{\CSLIndent}[1]{\hspace{\cslhangindent}#1}
$endif$
$if(lang)$
\ifLuaTeX
\usepackage[bidi=basic]{babel}
\else
\usepackage[bidi=default]{babel}
\fi
$if(babel-lang)$
\babelprovide[main,import]{$babel-lang$}
$if(mainfont)$
\ifPDFTeX
\else
\babelfont[$babel-lang$]{rm}{$mainfont$}
\fi
$endif$
$endif$
$for(babel-otherlangs)$
\babelprovide[import]{$babel-otherlangs$}
$endfor$
$for(babelfonts)$
\babelfont[$babelfonts.language$]{rm}{$babelfonts.font$}
$endfor$
% get rid of language-specific shorthands (see #6817):
\let\LanguageShortHands\languageshorthands
\def\languageshorthands#1{}
$endif$
$for(header-includes)$
$header-includes$
$endfor$
\ifLuaTeX
  \usepackage{selnolig}  % disable illegal ligatures
\fi
$if(dir)$
\ifPDFTeX
  \TeXXeTstate=1
  \newcommand{\RL}[1]{\beginR #1\endR}
  \newcommand{\LR}[1]{\beginL #1\endL}
  \newenvironment{RTL}{\beginR}{\endR}
  \newenvironment{LTR}{\beginL}{\endL}
\fi
$endif$
$if(natbib)$
\usepackage[$natbiboptions$]{natbib}
\bibliographystyle{$if(biblio-style)$$biblio-style$$else$plainnat$endif$}
$endif$
$if(biblatex)$
\usepackage[$if(biblio-style)$style=$biblio-style$,$endif$$for(biblatexoptions)$$biblatexoptions$$sep$,$endfor$]{biblatex}
$for(bibliography)$
\addbibresource{$bibliography$}
$endfor$
$endif$
$if(nocite-ids)$
\nocite{$for(nocite-ids)$$it$$sep$, $endfor$}
$endif$
$if(csquotes)$
\usepackage{csquotes}
$endif$
\IfFileExists{bookmark.sty}{\usepackage{bookmark}}{\usepackage{hyperref}}
\IfFileExists{xurl.sty}{\usepackage{xurl}}{} % add URL line breaks if available
\urlstyle{$if(urlstyle)$$urlstyle$$else$same$endif$}
$if(links-as-notes)$
% Make links footnotes instead of hotlinks:
\DeclareRobustCommand{\href}[2]{#2\footnote{\url{#1}}}
$endif$
$if(verbatim-in-note)$
\VerbatimFootnotes % allow verbatim text in footnotes
$endif$
\hypersetup{
$if(title-meta)$
  pdftitle={$title-meta$},
$endif$
$if(author-meta)$
  pdfauthor={$author-meta$},
$endif$
$if(lang)$
  pdflang={$lang$},
$endif$
$if(subject)$
  pdfsubject={$subject$},
$endif$
$if(keywords)$
  pdfkeywords={$for(keywords)$$keywords$$sep$, $endfor$},
$endif$
$if(colorlinks)$
  colorlinks=true,
  linkcolor={$if(linkcolor)$$linkcolor$$else$default-linkcolor$endif$},
  filecolor={$if(filecolor)$$filecolor$$else$default-filecolor$endif$},
  citecolor={$if(citecolor)$$citecolor$$else$default-citecolor$endif$},
  urlcolor={$if(urlcolor)$$urlcolor$$else$default-urlcolor$endif$},
$else$
  hidelinks,
$endif$
  breaklinks=true,
  pdfcreator={LaTeX via pandoc with the Eisvogel template}}
$if(title)$
\title{$title$$if(thanks)$\thanks{$thanks$}$endif$}
$endif$
$if(subtitle)$
$if(beamer)$
$else$
\usepackage{etoolbox}
\makeatletter
\providecommand{\subtitle}[1]{% add subtitle to \maketitle
  \apptocmd{\@title}{\par {\large #1 \par}}{}{}
}
\makeatother
$endif$
\subtitle{$subtitle$}
$endif$
\author{$for(author)$$author$$sep$ \and $endfor$}
\date{$date$}
$if(beamer)$
$if(institute)$
\institute{$for(institute)$$institute$$sep$ \and $endfor$}
$endif$
$if(titlegraphic)$
\titlegraphic{\includegraphics{$titlegraphic$}}
$endif$
$if(logo)$
\logo{\includegraphics{$logo$}}
$endif$
$endif$



%%
%% added
%%

$if(page-background)$
\usepackage[pages=all]{background}
$endif$

%
% for the background color of the title page
%
$if(titlepage)$
\usepackage{pagecolor}
\usepackage{afterpage}
$if(titlepage-background)$
\usepackage{tikz}
$endif$
$if(geometry)$
$else$
\usepackage[margin=2.5cm,includehead=true,includefoot=true,centering]{geometry}
$endif$
$endif$

%
% break urls
%
\PassOptionsToPackage{hyphens}{url}

%
% When using babel or polyglossia with biblatex, loading csquotes is recommended
% to ensure that quoted texts are typeset according to the rules of your main language.
%
\usepackage{csquotes}

%
% captions
%
\definecolor{caption-color}{HTML}{777777}
$if(beamer)$
$else$
\usepackage[font={stretch=1.2}, textfont={color=caption-color}, position=top, skip=4mm, labelfont=bf, singlelinecheck=false, justification=$if(caption-justification)$$caption-justification$$else$raggedright$endif$]{caption}
\setcapindent{0em}
$endif$

%
% blockquote
%
\definecolor{blockquote-border}{RGB}{221,221,221}
\definecolor{blockquote-text}{RGB}{119,119,119}
\usepackage{mdframed}
\newmdenv[rightline=false,bottomline=false,topline=false,linewidth=3pt,linecolor=blockquote-border,skipabove=\parskip]{customblockquote}
\renewenvironment{quote}{\begin{customblockquote}\list{}{\rightmargin=0em\leftmargin=0em}%
\item\relax\color{blockquote-text}\ignorespaces}{\unskip\unskip\endlist\end{customblockquote}}

%
% Source Sans Pro as the de­fault font fam­ily
% Source Code Pro for monospace text
%
% 'default' option sets the default
% font family to Source Sans Pro, not \sfdefault.
%
\ifnum 0\ifxetex 1\fi\ifluatex 1\fi=0 % if pdftex
  $if(fontfamily)$
  $else$
  \usepackage[default]{sourcesanspro}
  \usepackage{sourcecodepro}
  $endif$
\else % if not pdftex
  $if(mainfont)$
  $else$
  \usepackage[default]{sourcesanspro}
  \usepackage{sourcecodepro}

  % XeLaTeX specific adjustments for straight quotes: https://tex.stackexchange.com/a/354887
  % This issue is already fixed (see https://github.com/silkeh/latex-sourcecodepro/pull/5) but the
  % fix is still unreleased.
  % TODO: Remove this workaround when the new version of sourcecodepro is released on CTAN.
  \ifxetex
    \makeatletter
    \defaultfontfeatures[\ttfamily]
      { Numbers   = \sourcecodepro@figurestyle,
        Scale     = \SourceCodePro@scale,
        Extension = .otf }
    \setmonofont
      [ UprightFont    = *-\sourcecodepro@regstyle,
        ItalicFont     = *-\sourcecodepro@regstyle It,
        BoldFont       = *-\sourcecodepro@boldstyle,
        BoldItalicFont = *-\sourcecodepro@boldstyle It ]
      {SourceCodePro}
    \makeatother
  \fi
  $endif$
\fi

%
% heading color
%
\definecolor{heading-color}{RGB}{40,40,40}
$if(beamer)$
$else$
\addtokomafont{section}{\color{heading-color}}
$endif$
% When using the classes report, scrreprt, book,
% scrbook or memoir, uncomment the following line.
%\addtokomafont{chapter}{\color{heading-color}}

%
% variables for title, author and date
%
$if(beamer)$
$else$
\usepackage{titling}
\title{$title$}
\author{$for(author)$$author$$sep$, $endfor$}
\date{$date$}
$endif$

%
% tables
%
$if(tables)$

\definecolor{table-row-color}{HTML}{F5F5F5}
\definecolor{table-rule-color}{HTML}{999999}

%\arrayrulecolor{black!40}
\arrayrulecolor{table-rule-color}     % color of \toprule, \midrule, \bottomrule
\setlength\heavyrulewidth{0.3ex}      % thickness of \toprule, \bottomrule
\renewcommand{\arraystretch}{1.3}     % spacing (padding)

$if(table-use-row-colors)$
% TODO: This doesn't work anymore. I don't know why.
% Reset rownum counter so that each table
% starts with the same row colors.
% https://tex.stackexchange.com/questions/170637/restarting-rowcolors
%
% Unfortunately the colored cells extend beyond the edge of the
% table because pandoc uses @-expressions (@{}) like so:
%
% \begin{longtable}[]{@{}ll@{}}
% \end{longtable}
%
% https://en.wikibooks.org/wiki/LaTeX/Tables#.40-expressions
\let\oldlongtable\longtable
\let\endoldlongtable\endlongtable
\renewenvironment{longtable}{
\rowcolors{3}{}{table-row-color!100}  % row color
\oldlongtable} {
\endoldlongtable
\global\rownum=0\relax}
$endif$
$endif$

%
% remove paragraph indention
%
\setlength{\parindent}{0pt}
\setlength{\parskip}{6pt plus 2pt minus 1pt}
\setlength{\emergencystretch}{3em}  % prevent overfull lines

%
%
% Listings
%
%

$if(listings)$

%
% general listing colors
%
\definecolor{listing-background}{HTML}{F7F7F7}
\definecolor{listing-rule}{HTML}{B3B2B3}
\definecolor{listing-numbers}{HTML}{B3B2B3}
\definecolor{listing-text-color}{HTML}{000000}
\definecolor{listing-keyword}{HTML}{435489}
\definecolor{listing-keyword-2}{HTML}{1284CA} % additional keywords
\definecolor{listing-keyword-3}{HTML}{9137CB} % additional keywords
\definecolor{listing-identifier}{HTML}{435489}
\definecolor{listing-string}{HTML}{00999A}
\definecolor{listing-comment}{HTML}{8E8E8E}

\lstdefinestyle{eisvogel_listing_style}{
  language         = java,
$if(listings-disable-line-numbers)$
  xleftmargin      = 0.6em,
  framexleftmargin = 0.4em,
$else$
  numbers          = left,
  xleftmargin      = 2.7em,
  framexleftmargin = 2.5em,
$endif$
  backgroundcolor  = \color{listing-background},
  basicstyle       = \color{listing-text-color}\linespread{1.0}%
                      \lst@ifdisplaystyle%
                      $if(code-block-font-size)$$code-block-font-size$$else$\small$endif$%
                      \fi\ttfamily{},
  breaklines       = true,
  frame            = single,
  framesep         = 0.19em,
  rulecolor        = \color{listing-rule},
  frameround       = ffff,
  tabsize          = 4,
  numberstyle      = \color{listing-numbers},
  aboveskip        = 1.0em,
  belowskip        = 0.1em,
  abovecaptionskip = 0em,
  belowcaptionskip = 1.0em,
  keywordstyle     = {\color{listing-keyword}\bfseries},
  keywordstyle     = {[2]\color{listing-keyword-2}\bfseries},
  keywordstyle     = {[3]\color{listing-keyword-3}\bfseries\itshape},
  sensitive        = true,
  identifierstyle  = \color{listing-identifier},
  commentstyle     = \color{listing-comment},
  stringstyle      = \color{listing-string},
  showstringspaces = false,
  escapeinside     = {/*@}{@*/}, % Allow LaTeX inside these special comments
  literate         =
  {á}{{\'a}}1 {é}{{\'e}}1 {í}{{\'i}}1 {ó}{{\'o}}1 {ú}{{\'u}}1
  {Á}{{\'A}}1 {É}{{\'E}}1 {Í}{{\'I}}1 {Ó}{{\'O}}1 {Ú}{{\'U}}1
  {à}{{\`a}}1 {è}{{\`e}}1 {ì}{{\`i}}1 {ò}{{\`o}}1 {ù}{{\`u}}1
  {À}{{\`A}}1 {È}{{\`E}}1 {Ì}{{\`I}}1 {Ò}{{\`O}}1 {Ù}{{\`U}}1
  {ä}{{\"a}}1 {ë}{{\"e}}1 {ï}{{\"i}}1 {ö}{{\"o}}1 {ü}{{\"u}}1
  {Ä}{{\"A}}1 {Ë}{{\"E}}1 {Ï}{{\"I}}1 {Ö}{{\"O}}1 {Ü}{{\"U}}1
  {â}{{\^a}}1 {ê}{{\^e}}1 {î}{{\^i}}1 {ô}{{\^o}}1 {û}{{\^u}}1
  {Â}{{\^A}}1 {Ê}{{\^E}}1 {Î}{{\^I}}1 {Ô}{{\^O}}1 {Û}{{\^U}}1
  {œ}{{\oe}}1 {Œ}{{\OE}}1 {æ}{{\ae}}1 {Æ}{{\AE}}1 {ß}{{\ss}}1
  {ç}{{\c c}}1 {Ç}{{\c C}}1 {ø}{{\o}}1 {å}{{\r a}}1 {Å}{{\r A}}1
  {€}{{\EUR}}1 {£}{{\pounds}}1 {«}{{\guillemotleft}}1
  {»}{{\guillemotright}}1 {ñ}{{\~n}}1 {Ñ}{{\~N}}1 {¿}{{?`}}1
  {…}{{\ldots}}1 {≥}{{>=}}1 {≤}{{<=}}1 {„}{{\glqq}}1 {“}{{\grqq}}1
  {”}{{''}}1
}
\lstset{style=eisvogel_listing_style}

%
% Java (Java SE 12, 2019-06-22)
%
\lstdefinelanguage{Java}{
  morekeywords={
    % normal keywords (without data types)
    abstract,assert,break,case,catch,class,continue,default,
    do,else,enum,exports,extends,final,finally,for,if,implements,
    import,instanceof,interface,module,native,new,package,private,
    protected,public,requires,return,static,strictfp,super,switch,
    synchronized,this,throw,throws,transient,try,volatile,while,
    % var is an identifier
    var
  },
  morekeywords={[2] % data types
    % primitive data types
    boolean,byte,char,double,float,int,long,short,
    % String
    String,
    % primitive wrapper types
    Boolean,Byte,Character,Double,Float,Integer,Long,Short
    % number types
    Number,AtomicInteger,AtomicLong,BigDecimal,BigInteger,DoubleAccumulator,DoubleAdder,LongAccumulator,LongAdder,Short,
    % other
    Object,Void,void
  },
  morekeywords={[3] % literals
    % reserved words for literal values
    null,true,false,
  },
  sensitive,
  morecomment  = [l]//,
  morecomment  = [s]{/*}{*/},
  morecomment  = [s]{/**}{*/},
  morestring   = [b]",
  morestring   = [b]',
}

\lstdefinelanguage{XML}{
  morestring      = [b]",
  moredelim       = [s][\bfseries\color{listing-keyword}]{<}{\ },
  moredelim       = [s][\bfseries\color{listing-keyword}]{</}{>},
  moredelim       = [l][\bfseries\color{listing-keyword}]{/>},
  moredelim       = [l][\bfseries\color{listing-keyword}]{>},
  morecomment     = [s]{<?}{?>},
  morecomment     = [s]{<!--}{-->},
  commentstyle    = \color{listing-comment},
  stringstyle     = \color{listing-string},
  identifierstyle = \color{listing-identifier}
}
$endif$

%
% header and footer
%
$if(beamer)$
$else$
$if(disable-header-and-footer)$
$else$
\usepackage[headsepline,footsepline]{scrlayer-scrpage}

\newpairofpagestyles{eisvogel-header-footer}{
  \clearpairofpagestyles
  \ihead*{$if(header-left)$$header-left$$else$$title$$endif$}
  \chead*{$if(header-center)$$header-center$$else$$endif$}
  \ohead*{$if(header-right)$$header-right$$else$$date$$endif$}
  \ifoot*{$if(footer-left)$$footer-left$$else$$for(author)$$author$$sep$, $endfor$$endif$}
  \cfoot*{$if(footer-center)$$footer-center$$else$$endif$}
  \ofoot*{$if(footer-right)$$footer-right$$else$\thepage$endif$}
  \addtokomafont{pageheadfoot}{\upshape}
}
\pagestyle{eisvogel-header-footer}

$if(book)$
\deftripstyle{ChapterStyle}{}{}{}{}{\pagemark}{}
\renewcommand*{\chapterpagestyle}{ChapterStyle}
$endif$

$if(page-background)$
\backgroundsetup{
scale=1,
color=black,
opacity=$if(page-background-opacity)$$page-background-opacity$$else$0.2$endif$,
angle=0,
contents={%
  \includegraphics[width=\paperwidth,height=\paperheight]{$page-background$}
  }%
}
$endif$
$endif$
$endif$

%%
%% end added
%%

\begin{document}

%%
%% begin titlepage
%%
$if(beamer)$
$else$
$if(titlepage)$
\begin{titlepage}
$if(titlepage-background)$
\newgeometry{top=2cm, right=4cm, bottom=3cm, left=4cm}
$else$
\newgeometry{left=6cm}
$endif$
$if(titlepage-color)$
\definecolor{titlepage-color}{HTML}{$titlepage-color$}
\newpagecolor{titlepage-color}\afterpage{\restorepagecolor}
$endif$
$if(titlepage-background)$
\tikz[remember picture,overlay] \node[inner sep=0pt] at (current page.center){\includegraphics[width=\paperwidth,height=\paperheight]{$titlepage-background$}};
$endif$
\newcommand{\colorRule}[3][black]{\textcolor[HTML]{#1}{\rule{#2}{#3}}}
\begin{flushleft}
\noindent
\\[-1em]
\color[HTML]{$if(titlepage-text-color)$$titlepage-text-color$$else$5F5F5F$endif$}
\makebox[0pt][l]{\colorRule[$if(titlepage-rule-color)$$titlepage-rule-color$$else$435488$endif$]{1.3\textwidth}{$if(titlepage-rule-height)$$titlepage-rule-height$$else$4$endif$pt}}
\par
\noindent

$if(titlepage-background)$
% The titlepage with a background image has other text spacing and text size
{
  \setstretch{2}
  \vfill
  \vskip -8em
  \noindent {\huge \textbf{\textsf{$title$}}}
  $if(subtitle)$
  \vskip 1em
  {\Large \textsf{$subtitle$}}
  $endif$
  \vskip 2em
  \noindent {\Large \textsf{$for(author)$$author$$sep$, $endfor$} \vskip 0.6em \textsf{$date$}}
  \vfill
}
$else$
{
  \setstretch{1.4}
  \vfill
  \noindent {\huge \textbf{\textsf{$title$}}}
  $if(subtitle)$
  \vskip 1em
  {\Large \textsf{$subtitle$}}
  $endif$
  \vskip 2em
  \noindent {\Large \textsf{$for(author)$$author$$sep$, $endfor$}}
  \vfill
}
$endif$

$if(titlepage-logo)$
\noindent
\includegraphics[width=$if(logo-width)$$logo-width$$else$35mm$endif$, left]{$titlepage-logo$}
$endif$

$if(titlepage-background)$
$else$
\textsf{$date$}
$endif$
\end{flushleft}
\end{titlepage}
\restoregeometry
\pagenumbering{arabic}
$endif$
$endif$

%%
%% end titlepage
%%

$if(has-frontmatter)$
\frontmatter
$endif$
$if(title)$
$if(beamer)$
\frame{\titlepage}
% don't generate the default title
% $else$
% \maketitle
$endif$
$if(abstract)$
\begin{abstract}
$abstract$
\end{abstract}
$endif$
$endif$

$if(first-chapter)$
\setcounter{chapter}{$first-chapter$}
\addtocounter{chapter}{-1}
$endif$

$for(include-before)$
$include-before$

$endfor$
$if(toc)$
$if(toc-title)$
\renewcommand*\contentsname{$toc-title$}
$endif$
$if(beamer)$
\begin{frame}[allowframebreaks]
$if(toc-title)$
  \frametitle{$toc-title$}
$endif$
  \tableofcontents[hideallsubsections]
\end{frame}
$if(toc-own-page)$
\newpage
$endif$
$else$
{
$if(colorlinks)$
\hypersetup{linkcolor=$if(toccolor)$$toccolor$$else$$endif$}
$endif$
\setcounter{tocdepth}{$toc-depth$}
\tableofcontents
$if(toc-own-page)$
\newpage
$endif$
}
$endif$
$endif$
$if(lof)$
\listoffigures
$endif$
$if(lot)$
\listoftables
$endif$
$if(linestretch)$
\setstretch{$linestretch$}
$endif$
$if(has-frontmatter)$
\mainmatter
$endif$
$body$

$if(has-frontmatter)$
\backmatter
$endif$
$if(natbib)$
$if(bibliography)$
$if(biblio-title)$
$if(has-chapters)$
\renewcommand\bibname{$biblio-title$}
$else$
\renewcommand\refname{$biblio-title$}
$endif$
$endif$
$if(beamer)$
\begin{frame}[allowframebreaks]{$biblio-title$}
  \bibliographytrue
$endif$
  \bibliography{$for(bibliography)$$bibliography$$sep$,$endfor$}
$if(beamer)$
\end{frame}
$endif$

$endif$
$endif$
$if(biblatex)$
$if(beamer)$
\begin{frame}[allowframebreaks]{$biblio-title$}
  \bibliographytrue
  \printbibliography[heading=none]
\end{frame}
$else$
\printbibliography$if(biblio-title)$[title=$biblio-title$]$endif$
$endif$

$endif$
$for(include-after)$
$include-after$

$endfor$
\end{document}
"""


def load_eis():
    if not GLOBAL_PATH.parent.exists():
        GLOBAL_PATH.parent.mkdir(parents=True)

    eis_path = GLOBAL_PATH.parent / "eisvogel.tex"

    if not eis_path.exists():
        with open(eis_path, "w") as f:
            f.write(EISVOGEL)

    return eis_path


async def aload_eis():
    if not GLOBAL_PATH.parent.exists():
        GLOBAL_PATH.parent.mkdir(parents=True)

    eis_path = GLOBAL_PATH.parent / "eisvogel.tex"

    if not eis_path.exists():
        async with aiofiles.open(eis_path, "w") as f:
            await f.write(EISVOGEL)
    return eis_path



import argparse
import sys
from typing import Any, Optional

from loguru import logger

from .commands import (
    chat_async_biorxiv,
    chat_async_paper,
    chat_config,
    chat_response,
    chat_reviewer,
)


class RichArgParser(argparse.ArgumentParser):
    """RichArgParser."""

    def __init__(self, *args: Any, **kwargs: Any):
        """RichArgParser."""
        from rich.console import Console

        self.console = Console()
        super().__init__(*args, **kwargs)

    @staticmethod
    def _color_message(message: str, color: str = "green") -> str:
        """Color message."""
        import re

        pattern = re.compile(r"(?P<arg>-{1,2}[-|\w]+)")
        return pattern.sub(lambda m: f"[bold {color}]{m.group('arg')}[/]", message)

    def _print_message(self, message: Optional[str], file: Any = None) -> None:
        if message:
            self.console.print(self._color_message(message))


class RichHelpFormatter(argparse.RawDescriptionHelpFormatter):
    """RichHelpFormatter."""

    def __init__(self, *args: Any, **kwargs: Any):
        """RichHelpFormatter."""
        super().__init__(*args, max_help_position=42, width=100, **kwargs)  # type: ignore


def _cli():
    parser = RichArgParser(
        description="[red]chatre[/] Use ChatGPT to accelerate research",
        formatter_class=RichHelpFormatter,
        epilog="""example usage:
    Read one paper from local path and export result as md
        chatre paper --pdf example.pdf
    Read one paper from local path and export result as pdf
        chatre paper --pdf example.pdf --file-format pdf
    Reader papers from local directory recursively
        chatre paper --pdf example/
    """,
    )

    parser.add_argument(
        "-l",
        "--log-level",
        type=str,
        default="info",
        choices=["debug", "info", "trace"],
        metavar="[debug, info, trace]",
        help="The log level (default: %(default)s)",
    )

    subparser = parser.add_subparsers(
        title="subcommand",
        description="valid subcommand",
        dest="subcommand",
    )

    chat_reviewer_command = chat_reviewer.add_subcommand(subparser)
    chat_response_command = chat_response.add_subcommand(subparser)
    chat_config_command = chat_config.add_subcommand(subparser)
    chat_async_biorxiv_command = chat_async_biorxiv.add_subcommand(subparser)
    chat_async_paper_command = chat_async_paper.add_subcommand(subparser)

    args = parser.parse_args()
    debug_format = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>"
    logger.remove()

    if args.log_level.upper() == "INFO":
        logger.add(
            sys.stdout,
            format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <level>{message}</level>",
            level=args.log_level.upper(),
        )
    else:
        logger.add(
            sys.stdout,
            format=debug_format,
            level=args.log_level.upper(),
        )

    if not args.subcommand:
        parser.print_help()
        raise SystemExit

    if args.subcommand == chat_response_command:
        chat_response.cli(args)
    elif args.subcommand == chat_reviewer_command:
        chat_reviewer.cli(args)
    elif args.subcommand == chat_config_command:
        chat_config.cli(args)
    elif args.subcommand == chat_async_biorxiv_command:
        chat_async_biorxiv.cli(args)
    elif args.subcommand == chat_async_paper_command:
        chat_async_paper.cli(args)
    else:
        logger.error("Invalid subcommand")
        parser.print_help()
        raise SystemExit

    logger.success("Done")


def cli():
    try:
        _cli()
    except KeyboardInterrupt:
        logger.info("Say youn again!")
        raise SystemExit



import json
import re
from collections import Counter

import fitz


class Paper:
    def __init__(self, path, title="", url="", abs="", authors=[]):
        # 初始化函数，根据pdf路径初始化Paper对象
        self.url = url  # 文章链接
        self.path = path  # pdf路径
        self.section_names = []  # 段落标题
        self.section_texts = {}  # 段落内容
        self.abs = abs
        self.title_page = 0

        if title == "":
            self.pdf = fitz.open(self.path)  # pdf文档
            self.title = self.get_title()
            self.parse_pdf()
        else:
            self.title = title

        self.authors = authors
        self.roman_num = [
            "I",
            "II",
            "III",
            "IV",
            "V",
            "VI",
            "VII",
            "VIII",
            "IIX",
            "IX",
            "X",
        ]
        self.digit_num = [str(d + 1) for d in range(10)]
        self.first_image = ""

    def parse_pdf(self):
        self.pdf = fitz.open(self.path)  # pdf文档
        self.text_list = [page.get_text() for page in self.pdf]
        self.all_text = " ".join(self.text_list)
        self.extract_section_infomation()
        self.section_texts.update({"title": self.title})
        self.pdf.close()

    # 定义一个函数，根据字体的大小，识别每个章节名称，并返回一个列表
    def get_chapter_names(
        self,
    ):
        # # 打开一个pdf文件
        doc = fitz.open(self.path)  # pdf文档
        text_list = [page.get_text() for page in doc]
        all_text = ""
        for text in text_list:
            all_text += text
        # # 创建一个空列表，用于存储章节名称
        chapter_names = []
        for line in all_text.split("\n"):
            line.split(" ")
            if "." in line:
                point_split_list = line.split(".")
                space_split_list = line.split(" ")
                if 1 < len(space_split_list) < 5:
                    if 1 < len(point_split_list) < 5 and (
                        point_split_list[0] in self.roman_num
                        or point_split_list[0] in self.digit_num
                    ):
                        chapter_names.append(line)

        return chapter_names

    def fetch_title(self):
        return self.pdf.metadata["title"]

    def get_title(self):
        doc = self.pdf  # 打开pdf文件
        max_font_size = 0  # 初始化最大字体大小为0
        max_font_sizes = [0]
        for page_index, page in enumerate(doc):  # 遍历每一页
            text = page.get_text("dict")  # 获取页面上的文本信息
            blocks = text["blocks"]  # 获取文本块列表
            for block in blocks:  # 遍历每个文本块
                if block["type"] == 0 and len(block["lines"]):  # 如果是文字类型
                    if len(block["lines"][0]["spans"]):
                        font_size = block["lines"][0]["spans"][0][
                            "size"
                        ]  # 获取第一行第一段文字的字体大小
                        max_font_sizes.append(font_size)
                        if font_size > max_font_size:  # 如果字体大小大于当前最大值
                            max_font_size = font_size  # 更新最大值
                            block["lines"][0]["spans"][0]["text"]  # 更新最大值对应的字符串
        max_font_sizes.sort()
        cur_title = ""
        for page_index, page in enumerate(doc):  # 遍历每一页
            text = page.get_text("dict")  # 获取页面上的文本信息
            blocks = text["blocks"]  # 获取文本块列表
            for block in blocks:  # 遍历每个文本块
                if block["type"] == 0 and len(block["lines"]):  # 如果是文字类型
                    if len(block["lines"][0]["spans"]):
                        cur_string = block["lines"][0]["spans"][0][
                            "text"
                        ]  # 更新最大值对应的字符串
                        block["lines"][0]["spans"][0]["flags"]  # 获取第一行第一段文字的字体特征
                        font_size = block["lines"][0]["spans"][0][
                            "size"
                        ]  # 获取第一行第一段文字的字体大小
                        # print(font_size)
                        if (
                            abs(font_size - max_font_sizes[-1]) < 0.3
                            or abs(font_size - max_font_sizes[-2]) < 0.3
                        ):
                            # print("The string is bold.", max_string, "font_size:", font_size, "font_flags:", font_flags)
                            if len(cur_string) > 4 and "arXiv" not in cur_string:
                                # print("The string is bold.", max_string, "font_size:", font_size, "font_flags:", font_flags)
                                if cur_title == "":
                                    cur_title += cur_string
                                else:
                                    cur_title += " " + cur_string
                                self.title_page = page_index
                                # break
        title = cur_title.replace("\n", " ")
        return title

    def extract_section_infomation(self):
        doc = fitz.open(self.path)

        # 获取文档中所有字体大小
        font_sizes = []
        for page in doc:
            blocks = page.get_text("dict")["blocks"]
            for block in blocks:
                if "lines" not in block:
                    continue
                lines = block["lines"]
                for line in lines:
                    for span in line["spans"]:
                        font_sizes.append(span["size"])
        most_common_size, _ = Counter(font_sizes).most_common(1)[0]

        # 按照最频繁的字体大小确定标题字体大小的阈值
        threshold = most_common_size * 1
        section_dict = {}
        section_dict["Abstract"] = ""
        last_heading = None
        subheadings = []
        heading_font = -1
        # 遍历每一页并查找子标题
        found_abstract = False
        upper_heading = False
        font_heading = False
        for page in doc:
            blocks = page.get_text("dict")["blocks"]
            for block in blocks:
                if not found_abstract:
                    try:
                        text = json.dumps(block)
                    except Exception:
                        continue
                    if re.search(r"\bAbstract\b", text, re.IGNORECASE):
                        found_abstract = True
                        last_heading = "Abstract"
                if found_abstract:
                    if "lines" not in block:
                        continue
                    lines = block["lines"]
                    for line in lines:
                        for span in line["spans"]:
                            # 如果当前文本是子标题
                            if (
                                not font_heading
                                and span["text"].isupper()
                                and sum(
                                    1
                                    for c in span["text"]
                                    if c.isupper() and ("A" <= c <= "Z")
                                )
                                > 4
                            ):  # 针对一些标题大小一样,但是全大写的论文
                                upper_heading = True
                                heading = span["text"].strip()
                                if "References" in heading:  # reference 以后的内容不考虑
                                    self.section_names = subheadings
                                    self.section_texts = section_dict
                                    return
                                subheadings.append(heading)
                                if last_heading is not None:
                                    section_dict[last_heading] = section_dict[
                                        last_heading
                                    ].strip()
                                section_dict[heading] = ""
                                last_heading = heading
                            if (
                                not upper_heading
                                and span["size"] > threshold
                                and re.match(  # 正常情况下,通过字体大小判断
                                    r"[A-Z][a-z]+(?:\s[A-Z][a-z]+)*",
                                    span["text"].strip(),
                                )
                            ):
                                font_heading = True
                                if heading_font == -1:
                                    heading_font = span["size"]
                                elif heading_font != span["size"]:
                                    continue
                                heading = span["text"].strip()
                                if "References" in heading:  # reference 以后的内容不考虑
                                    self.section_names = subheadings
                                    self.section_texts = section_dict
                                    return
                                subheadings.append(heading)
                                if last_heading is not None:
                                    section_dict[last_heading] = section_dict[
                                        last_heading
                                    ].strip()
                                section_dict[heading] = ""
                                last_heading = heading
                            # 否则将当前文本添加到上一个子标题的文本中
                            elif last_heading is not None:
                                section_dict[last_heading] += " " + span["text"].strip()
        self.section_names = subheadings
        self.section_texts = section_dict



import asyncio
import base64
import datetime
import re
from pathlib import Path

import openai
import requests
import tenacity
import tiktoken
from loguru import logger

from .aexport import aexport
from .paper_with_image import Paper
from .utils import load_config


class AsyncBaseReader:
    def __init__(self, root_path, language, file_format, save_image):
        if isinstance(root_path, str):
            root_path = Path(root_path)

        self.root_path = root_path
        self.language = language
        self.file_format = file_format

        self.config, self.chat_api_list = load_config()
        self.cur_api = 0

        self.gitee_key = self.config["Gitee"]["api"] if save_image else ""

        self.max_token_num = 4096
        self.encoding = tiktoken.get_encoding("gpt2")
        self.token_usage = 0

    async def _summary_with_chat(self, paper_list: list[Paper], key_words):
        await asyncio.gather(
            *[
                self.summary_with_chat_for_one_paper(paper, index, key_words)
                for index, paper in enumerate(paper_list)
            ]
        )

    def summary_with_chat(self, paper_list: list[Paper], key_words):
        asyncio.run(self._summary_with_chat(paper_list, key_words))

    def update_title(self, text):
        for line in text.split("\n"):
            if "Title:" in line:
                return line.split("Title:")[1].strip()

    async def summary_with_chat_for_one_paper(
        self, paper: Paper, paper_index: int, key_words
    ):
        htmls = []
        text = ""
        text += "Title:" + paper.title
        text += "Url:" + paper.url
        text += "Abstrat:" + paper.abs
        text += "Paper_info:" + paper.section_text_dict["paper_info"]
        # intro
        text += list(paper.section_text_dict.values())[0]
        chat_summary_text = ""

        try:
            chat_summary_text = await self.chat_summary(text=text, key_words=key_words)
        except Exception as e:
            logger.warning(f"summary_error: {e}")
            if "maximum context" in str(e):
                current_tokens_index = (
                    str(e).find("your messages resulted in")
                    + len("your messages resulted in")
                    + 1
                )
                offset = int(str(e)[current_tokens_index : current_tokens_index + 4])
                summary_prompt_token = offset + 1000 + 150
                chat_summary_text = await self.chat_summary(
                    text=text,
                    key_words=key_words,
                    summary_prompt_token=summary_prompt_token,
                )
            else:
                raise e

        if (
            paper.title == ""
            and (title := self.update_title(chat_summary_text)) is not None
        ):
            paper.title = title

        htmls.append("## Paper:" + str(paper_index + 1))
        htmls.append("\n\n\n")
        htmls.append(chat_summary_text)

        # 第二步总结方法：
        # WARNING，由于有些文章的方法章节名是算法名，
        # 所以简单的通过关键词来筛选，很难获取，后面需要用其他的方案去优化。
        method_key = ""
        for parse_key in paper.section_text_dict.keys():
            if "method" in parse_key.lower() or "approach" in parse_key.lower():
                method_key = parse_key
                break

        if method_key != "":
            text = ""
            method_text = ""
            summary_text = ""
            summary_text += "<summary>" + chat_summary_text
            # methods
            method_text += paper.section_text_dict[method_key]
            text = summary_text + "\n\n<Methods>:\n\n" + method_text
            chat_method_text = ""
            try:
                chat_method_text = await self.chat_method(
                    text=text, key_words=key_words
                )
            except Exception as e:
                logger.error(f"method_error: {e}")
                if "maximum context" in str(e):
                    current_tokens_index = (
                        str(e).find("your messages resulted in")
                        + len("your messages resulted in")
                        + 1
                    )
                    offset = int(
                        str(e)[current_tokens_index : current_tokens_index + 4]
                    )
                    method_prompt_token = offset + 800 + 150
                    chat_method_text = await self.chat_method(
                        text=text,
                        key_words=key_words,
                        method_prompt_token=method_prompt_token,
                    )
            htmls.append(chat_method_text)
        else:
            chat_method_text = ""

        htmls.append("\n" * 4)

        # 第三步总结全文，并打分：
        conclusion_key = ""
        for parse_key in paper.section_text_dict.keys():
            if "conclu" in parse_key.lower():
                conclusion_key = parse_key
                break

        text = ""
        conclusion_text = ""
        summary_text = ""
        summary_text += (
            "<summary>"
            + chat_summary_text
            + "\n <Method summary>:\n"
            + chat_method_text
        )

        if conclusion_key != "":
            # conclusion
            conclusion_text += paper.section_text_dict[conclusion_key]
            text = summary_text + "\n\n<Conclusion>:\n\n" + conclusion_text
        else:
            text = summary_text

        chat_conclusion_text = ""
        try:
            chat_conclusion_text = await self.chat_conclusion(
                text=text, key_words=key_words
            )
        except Exception as e:
            logger.info(f"conclusion_error: {e}")
            if "maximum context" in str(e):
                current_tokens_index = (
                    str(e).find("your messages resulted in")
                    + len("your messages resulted in")
                    + 1
                )
                offset = int(str(e)[current_tokens_index : current_tokens_index + 4])
                conclusion_prompt_token = offset + 800 + 150
                chat_conclusion_text = await self.chat_conclusion(
                    text=text,
                    key_words=key_words,
                    conclusion_prompt_token=conclusion_prompt_token,
                )
        htmls.append(chat_conclusion_text)
        htmls.append("\n" * 4)

        # # 整合成一个文件，打包保存下来。
        date_str = str(datetime.datetime.now())[:13].replace(" ", "-")
        export_path = self.root_path / "export"

        if not export_path.exists():
            export_path.mkdir(parents=True, exist_ok=True)

        file_name = (
            Path(export_path) / f"{date_str}-{self.validateTitle(paper.title[:80])}"
        )

        await aexport(
            content="\n".join([item.strip() for item in htmls]),
            file_name=file_name.with_suffix(f".{self.file_format}"),
        )

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    async def chat_conclusion(self, text, key_words, conclusion_prompt_token=800):
        openai.api_key = self.chat_api_list[self.cur_api]
        self.cur_api += 1
        self.cur_api = (
            0 if self.cur_api >= len(self.chat_api_list) - 1 else self.cur_api
        )
        text_token = len(self.encoding.encode(text))
        clip_text_index = int(
            len(text) * (self.max_token_num - conclusion_prompt_token) / text_token
        )
        clip_text = text[:clip_text_index]

        messages = [
            {
                "role": "system",
                "content": f"You are a reviewer in the field of [{key_words}] and you need to critically review this article",
            },
            # chatgpt 角色
            {
                "role": "assistant",
                "content": "This is the <summary> and <conclusion> part of an English literature, where <summary> you have already summarized, but <conclusion> part, I need your help to summarize the following questions:"
                + clip_text,
            },
            # 背景知识，可以参考OpenReview的审稿流程
            {
                "role": "user",
                "content": """
                 8. Make the following summary.Be sure to use {} answers (proper nouns need to be marked in English).
                    - (1):What is the significance of this piece of work?
                    - (2):Summarize the strengths and weaknesses of this article in three dimensions: innovation point, performance, and workload.
                    .......
                 Follow the format of the output later:
                 8. Conclusion: \n\n
                    - (1):xxx;\n
                    - (2):Innovation point: xxx; Performance: xxx; Workload: xxx;\n

                 Be sure to use {} answers (proper nouns need to be marked in English), statements as concise and academic as possible, do not repeat the content of the previous <summary>, the value of the use of the original numbers, be sure to strictly follow the format, the corresponding content output to xxx, in accordance with \n line feed, ....... means fill in according to the actual requirements, if not, you can not write.
                 """.format(
                    self.language, self.language
                ),
            },
        ]
        response = await openai.ChatCompletion.acreate(
            model="gpt-3.5-turbo",
            # prompt需要用英语替换，少占用token。
            messages=messages,
        )
        result = ""
        for choice in response.choices:
            result += choice.message.content

        result = self.format_text(result)
        logger.trace(f"conclusion_result:\n{result}")
        self.report_token_usage(response)

        return result

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    async def chat_method(self, text, key_words, method_prompt_token=800):
        openai.api_key = self.chat_api_list[self.cur_api]
        self.cur_api += 1
        self.cur_api = (
            0 if self.cur_api >= len(self.chat_api_list) - 1 else self.cur_api
        )
        text_token = len(self.encoding.encode(text))
        clip_text_index = int(
            len(text) * (self.max_token_num - method_prompt_token) / text_token
        )
        clip_text = text[:clip_text_index]

        messages = [
            {
                "role": "system",
                "content": f"You are a researcher in the field of [{key_words}] who is good at summarizing papers using concise statements",
            },
            # chatgpt 角色
            {
                "role": "assistant",
                "content": "This is the <summary> and <Method> part of an English document, where <summary> you have summarized, but the <Methods> part, I need your help to read and summarize the following questions."
                + clip_text,
            },
            # 背景知识
            {
                "role": "user",
                "content": """
                 7. Describe in detail the methodological idea of this article. Be sure to use {} answers (proper nouns need to be marked in English). For example, its steps are.
                    - (1):...
                    - (2):...
                    - (3):...
                    - .......
                 Follow the format of the output that follows:
                 7. Methods: \n\n
                    - (1):xxx;\n
                    - (2):xxx;\n
                    - (3):xxx;\n
                    ....... \n\n

                 Be sure to use {} answers (proper nouns need to be marked in English), statements as concise and academic as possible, do not repeat the content of the previous <summary>, the value of the use of the original numbers, be sure to strictly follow the format, the corresponding content output to xxx, in accordance with \n line feed, ....... means fill in according to the actual requirements, if not, you can not write.
                 """.format(
                    self.language, self.language
                ),
            },
        ]
        response = await openai.ChatCompletion.acreate(
            model="gpt-3.5-turbo",
            messages=messages,
        )

        result = ""
        for choice in response.choices:
            result += choice.message.content

        result = self.format_text(result)
        logger.trace(f"method_result:\n{result}")
        self.report_token_usage(response)

        return result

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    async def chat_summary(self, text, key_words, summary_prompt_token=1100):
        openai.api_key = self.chat_api_list[self.cur_api]
        self.cur_api += 1
        self.cur_api = (
            0 if self.cur_api >= len(self.chat_api_list) - 1 else self.cur_api
        )
        text_token = len(self.encoding.encode(text))
        clip_text_index = int(
            len(text) * (self.max_token_num - summary_prompt_token) / text_token
        )
        clip_text = text[:clip_text_index]

        messages = [
            {
                "role": "system",
                "content": f"You are a researcher in the field of [{key_words}]  who is good at summarizing papers using concise statements",
            },
            {
                "role": "assistant",
                "content": "This is the title, author, link, abstract and introduction of an English document. I need your help to read and summarize the following questions: "
                + clip_text,
            },
            {
                "role": "user",
                "content": f"""
                 1. Mark the title of the paper
                 2. list all the authors' names (use English)
                 3. mark the first author's affiliation (use English)
                 4. mark the keywords of this article (use English)
                 5. link to the paper, Github code link (if available, fill in Github:None if not)
                 6. summarize according to the following four points.Be sure to use {self.language} answers (proper nouns need to be marked in English)
                    - (1):What is the research background of this article?
                    - (2):What are the past methods? What are the problems with them? Is the approach well motivated?
                    - (3):What is the research methodology proposed in this paper?
                    - (4):On what task and what performance is achieved by the methods in this paper? Can the performance support their goals?
                 Follow the format of the output that follows:
                 1. Title: xxx\n\n
                 2. Authors: xxx\n\n
                 3. Affiliation: xxx\n\n
                 4. Keywords: xxx\n\n
                 5. Urls: xxx or xxx , xxx \n\n
                 6. Summary: \n\n
                    - (1):xxx;\n
                    - (2):xxx;\n
                    - (3):xxx;\n
                    - (4):xxx.\n\n

                 Be sure to use {self.language} answers (proper nouns need to be marked in English), statements as concise and academic as possible, do not have too much repetitive information, numerical values using the original numbers, be sure to strictly follow the format,
                 the corresponding content output to xxx, in accordance with \n line feed.
                 """,
            },
        ]

        response = await openai.ChatCompletion.acreate(
            model="gpt-3.5-turbo",
            messages=messages,
        )
        result = ""

        for choice in response.choices:
            result += choice.message.content

        result = self.format_text(result)
        logger.trace(f"summary_result:\n{result}")

        self.report_token_usage(response)

        return result

    @staticmethod
    def format_text(text):
        result = ""
        for line in text.split("\n"):
            result += line.strip() + "\n"
        return result

    def validateTitle(self, title):
        # 将论文的乱七八糟的路径格式修正
        rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
        new_title = re.sub(rstr, "_", title)  # 替换为下划线
        return new_title

    def show_token_usage(self):
        money = self.token_usage / 1000 * 0.002
        logger.info(f"TOKENS: {self.token_usage} / PRICES: ${money:.6f}")

    def report_token_usage(self, response):
        logger.trace(f"prompt_token_used: {response.usage.prompt_tokens}")
        logger.trace(f"completion_token_used: {response.usage.completion_tokens}")
        logger.trace(f"total_token_used: {response.usage.total_tokens}")
        logger.trace(f"response_time: { response.response_ms / 1000.0}s")
        self.token_usage += response.usage.total_tokens

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    def upload_gitee(self, image_path, image_name="", ext="png"):
        """
        上传到码云
        :return:
        """
        with open(image_path, "rb") as f:
            base64_data = base64.b64encode(f.read())
            base64_content = base64_data.decode()

        date_str = (
            str(datetime.datetime.now())[:19].replace(":", "-").replace(" ", "-")
            + "."
            + ext
        )
        path = image_name + "-" + date_str

        payload = {
            "access_token": self.gitee_key,
            "owner": self.config["Gitee"]["owner"],
            "repo": self.config["Gitee"]["repo"],
            "path": self.config["Gitee"]["path"],
            "content": base64_content,
            "message": "upload image",
        }
        # 这里需要修改成你的gitee的账户和仓库名，以及文件夹的名字：
        url = (
            "https://gitee.com/api/v5/repos/"
            + self.config["Gitee"]["owner"]
            + "/"
            + self.config["Gitee"]["repo"]
            + "/contents/"
            + self.config["Gitee"]["path"]
            + "/"
            + path
        )
        rep = requests.post(url, json=payload).json()
        logger.info(f"{rep=}")
        if "content" in rep.keys():
            image_url = rep["content"]["download_url"]
        else:
            image_url = (
                r"https://gitee.com/api/v5/repos/"
                + self.config["Gitee"]["owner"]
                + "/contents/"
                + self.config["Gitee"]["path"]
                + "/"
                + path
            )

        return image_url



# https://github.com/Wandmalfarbe/pandoc-latex-template
import asyncio
import shlex
import shutil
from pathlib import Path

import aiofiles
from loguru import logger

from .template import aload_eis


async def run(cmd) -> bool:
    proc = await asyncio.create_subprocess_shell(
        cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )

    stdout, stderr = await proc.communicate()

    logger.trace(f"[{cmd!r} exited with {proc.returncode}]")
    if stdout:
        logger.trace(f"[stdout]\n{stdout.decode()}")
    if stderr:
        logger.error(f"[stderr]\n{stderr.decode()}")

    return True if proc.returncode == 0 else False


async def acformt(file_name: Path, source: str, target: str) -> bool:
    from_file = file_name.with_suffix(f".{source}")
    to_file = file_name.with_suffix(f".{target}")

    if shutil.which("pandoc") is None:
        raise FileNotFoundError("pandoc not found")

    cmd = [
        "pandoc",
        from_file.as_posix(),
        "-o",
        to_file.as_posix(),
    ]

    _cmd = shlex.join(cmd)
    is_success = await run(_cmd)

    if not is_success:
        logger.warning("failed to output pdf then fall back to md")

    return is_success


async def amd2pdf(md_file: Path, pdf_file: Path, template=None) -> bool:
    if shutil.which("pandoc") is None:
        raise FileNotFoundError("pandoc not found")

    cmd = [
        "pandoc",
        md_file.as_posix(),
        "-o",
        pdf_file.as_posix(),
    ]

    if template is not None:
        if isinstance(template, Path):
            cmd += ["--template", template.as_posix()]
        else:
            cmd += ["--template", template]

    _cmd = shlex.join(cmd)
    is_success = await run(_cmd)

    if not is_success:
        logger.warning("failed to output pdf then fall back to md")

    return is_success


async def amd2tex(md_file: Path, tex_file: Path) -> bool:
    assert md_file.suffix == ".md"
    assert tex_file.suffix == ".tex"
    if shutil.which("pandoc") is None:
        raise FileNotFoundError("pandoc not found")

    cmd = [
        "pandoc",
        md_file.as_posix(),
        "-o",
        tex_file.as_posix(),
    ]

    _cmd = shlex.join(cmd)
    is_success = await run(_cmd)

    if not is_success:
        logger.warning("failed to output pdf then fall back to md")

    return is_success


async def aexport_to_markdown(text: str, file_name, mode="w"):
    async with aiofiles.open(file_name, mode, encoding="utf-8") as f:
        await f.write(text)


async def aexport_to_pdf(text: str, file_name: Path, mode="w", keep_md: bool = False):
    await aexport_to_markdown(text, file_name.with_suffix(".md"), mode)
    template = await aload_eis()
    flag = await amd2pdf(
        file_name.with_suffix(".md"), file_name.with_suffix(".pdf"), template
    )
    if flag:
        if not keep_md:
            file_name.with_suffix(".md").unlink()


async def aexport_to_tex(text: str, file_name: Path, mode="w", keep_md: bool = False):
    await aexport_to_markdown(text, file_name.with_suffix(".md"), mode)
    flag = await amd2tex(file_name.with_suffix(".md"), file_name.with_suffix(".tex"))

    if flag:
        if not keep_md:
            file_name.with_suffix(".md").unlink()


async def aexport(content: str, file_name: Path, keep_md: bool = False):
    if file_name.suffix == ".md" or file_name.suffix == ".txt":
        await aexport_to_markdown(content, file_name, mode="w")
    elif file_name.suffix == ".pdf":
        await aexport_to_pdf(content, file_name, mode="w", keep_md=keep_md)
    elif file_name.suffix == ".tex":
        await aexport_to_tex(content, file_name, mode="w", keep_md=keep_md)
    else:
        raise ValueError("Unsupported file format")



import datetime
import os
import re
from pathlib import Path
from typing import Optional

import openai
import tenacity
import tiktoken
from loguru import logger
from pydantic import BaseModel, validator

from chat_research.utils import report_token_usage

from ..paper import Paper
from ..utils import load_config


class ReviewerParams(BaseModel):
    paper_path: str
    file_format: str
    review_format: Optional[str] = None
    research_fields: str
    language: str

    @validator("paper_path")
    def paper_path_must_exist(cls, v):
        if not Path(v).exists():
            raise ValueError("paper_path must exist")


REVIEW_FORMAT = """
* Overall Review
Please briefly summarize the main points and contributions of this paper.
xxx

* Paper Strength
Please provide a list of the strengths of this paper, including but not limited to: innovative and practical methodology, insightful empirical findings or in-depth theoretical analysis, well-structured review of relevant literature, and any other factors that may make the paper valuable to readers. (Maximum length: 2,000 characters)
(1) xxx
(2) xxx
(3) xxx
...

* Paper Weakness
Please provide a numbered list of your main concerns regarding this paper (so authors could respond to the concerns individually). These may include, but are not limited to: inadequate implementation details for reproducing the study, limited evaluation and ablation studies for the proposed method, correctness of the theoretical analysis or experimental results, lack of comparisons or discussions with widely-known baselines in the field, lack of clarity in exposition, or any other factors that may impede the reader's understanding or benefit from the paper. Please kindly refrain from providing a general assessment of the paper's novelty without providing detailed explanations. (Maximum length: 2,000 characters)
(1) xxx
(2) xxx
(3) xxx
...

* Questions To Authors And Suggestions For Rebuttal
Please provide a numbered list of specific and clear questions that pertain to the details of the proposed method, evaluation setting, or additional results that would aid in supporting the authors' claims. The questions should be formulated in a manner that, after the authors have answered them during the rebuttal, it would enable a more thorough assessment of the paper's quality. (Maximum length: 2,000 characters)
xxx

*Overall score (1-10)
The paper is scored on a scale of 1-10, with 10 being the full mark, and 6 stands for borderline accept. Then give the reason for your rating.
xxx
"""


# 定义Reviewer类
class Reviewer:
    # 初始化方法，设置属性
    def __init__(self, args=None):
        if args is None:
            raise ValueError("args is None")

        if args.language == "en":
            self.language = "English"
        elif args.language == "zh":
            self.language = "Chinese"
        else:
            self.language = "English"

        self.research_fields = args.research_fields
        self.review_format = args.review_format

        self.config, self.chat_api_list = load_config()

        self.cur_api = 0
        self.file_format = args.file_format
        self.max_token_num = 4096
        self.encoding = tiktoken.get_encoding("gpt2")

    @staticmethod
    def get_review_format(path: Optional[Path]):
        if path is None:
            return REVIEW_FORMAT
        else:
            with open(path, "r") as f:
                return f.read()

    def validateTitle(self, title):
        # 修正论文的路径格式
        rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
        new_title = re.sub(rstr, "_", title)  # 替换为下划线
        return new_title

    def review_by_chatgpt(self, paper_list):
        htmls = []
        for paper_index, paper in enumerate(paper_list):
            sections_of_interest = self.stage_1(paper)
            # extract the essential parts of the paper
            text = ""
            text += "Title:" + paper.title + ". "
            text += "Abstract: " + paper.section_texts["Abstract"]
            intro_title = next(
                (item for item in paper.section_names if "ntroduction" in item.lower()),
                None,
            )
            if intro_title is not None:
                text += "Introduction: " + paper.section_texts[intro_title]
            # Similar for conclusion section
            conclusion_title = next(
                (item for item in paper.section_names if "onclusion" in item), None
            )
            if conclusion_title is not None:
                text += "Conclusion: " + paper.section_texts[conclusion_title]
            for heading in sections_of_interest:
                if heading in paper.section_names:
                    text += heading + ": " + paper.section_texts[heading]
            chat_review_text = self.chat_review(text=text)
            htmls.append("## Paper:" + str(paper_index + 1))
            htmls.append("\n\n\n")
            htmls.append(chat_review_text)

            # 将审稿意见保存起来
            date_str = str(datetime.datetime.now())[:13].replace(" ", "-")
            export_path = os.path.join(self.root_path, "export")
            if not os.path.exists(export_path):
                os.makedirs(export_path)
            mode = "w" if paper_index == 0 else "a"
            file_name = os.path.join(
                export_path,
                date_str
                + "-"
                + self.validateTitle(paper.title)
                + "."
                + self.file_format,
            )
            self.export_to_markdown("\n".join(htmls), file_name=file_name, mode=mode)
            htmls = []

    def stage_1(self, paper):
        text = ""
        text += "Title: " + paper.title + ". "
        text += "Abstract: " + paper.section_texts["Abstract"]
        openai.api_key = self.chat_api_list[self.cur_api]
        self.cur_api += 1
        self.cur_api = (
            0 if self.cur_api >= len(self.chat_api_list) - 1 else self.cur_api
        )
        messages = [
            {
                "role": "system",
                "content": f"You are a professional reviewer in the field of {self.research_fields}. "
                f"I will give you a paper. You need to review this paper and discuss the novelty and originality of ideas, correctness, clarity, the significance of results, potential impact and quality of the presentation. "
                f"Due to the length limitations, I am only allowed to provide you the abstract, introduction, conclusion and at most two sections of this paper."
                f"Now I will give you the title and abstract and the headings of potential sections. "
                f"You need to reply at most two headings. Then I will further provide you the full information, includes aforementioned sections and at most two sections you called for.\n\n"
                f"Title: {paper.title}\n\n"
                f"Abstract: {paper.section_texts['Abstract']}\n\n"
                f"Potential Sections: {paper.section_names[2:-1]}\n\n"
                f"Follow the following format to output your choice of sections:"
                f"{{chosen section 1}}, {{chosen section 2}}\n\n",
            },
            {"role": "user", "content": text},
        ]
        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=messages,
        )
        result = ""
        for choice in response.choices:
            result += choice.message.content
        logger.info(result)
        return result.split(",")

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
        stop=tenacity.stop_after_attempt(5),
        reraise=True,
    )
    def chat_review(self, text):
        openai.api_key = self.chat_api_list[self.cur_api]
        self.cur_api += 1
        self.cur_api = (
            0 if self.cur_api >= len(self.chat_api_list) - 1 else self.cur_api
        )
        review_prompt_token = 1000
        text_token = len(self.encoding.encode(text))
        input_text_index = int(
            len(text) * (self.max_token_num - review_prompt_token) / text_token
        )
        input_text = "This is the paper for your review:" + text[:input_text_index]

        review_format = self.get_review_format(self.review_format)

        messages = [
            {
                "role": "system",
                "content": "You are a professional reviewer in the field of "
                + self.research_fields
                + ". Now I will give you a paper. You need to give a complete review opinion according to the following requirements and format:"
                + review_format
                + " Please answer in {}.".format(self.language),
            },
            {"role": "user", "content": input_text},
        ]

        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=messages,
        )
        result = ""
        for choice in response.choices:
            result += choice.message.content
        logger.info("********" * 10)
        logger.info(result)
        logger.info("********" * 10)

        report_token_usage(response)

        return result

    def export_to_markdown(self, text, file_name, mode="w"):
        # 使用markdown模块的convert方法，将文本转换为html格式
        # html = markdown.markdown(text)
        # 打开一个文件，以写入模式
        with open(file_name, mode, encoding="utf-8") as f:
            # 将html格式的内容写入文件
            f.write(text)


def add_subcommand(parser):
    name = "reviewer"
    subparser = parser.add_parser(name, help="Summary paper")
    subparser.add_argument(
        "-p",
        "--paper-path",
        type=str,
        metavar="",
        help="path of papers",
        required=True,
    )

    subparser.add_argument(
        "-f",
        "--file-format",
        type=str,
        default="md",
        choices=["txt", "md"],
        metavar="",
        help="output file format (default: %(default)s)",
    )

    subparser.add_argument(
        "-r", "--review-format", type=str, metavar="", help="review format"
    )

    subparser.add_argument(
        "--research-fields",
        type=str,
        default="computer science, artificial intelligence and reinforcement learning",
        metavar="",
        help="the research fields of paper (default: %(default)s)",
    )

    subparser.add_argument(
        "-l",
        "--language",
        type=str,
        default="en",
        metavar="",
        help="output language, en or zh (default: %(default)s)",
    )

    return name


def main(args):
    reviewer1 = Reviewer(args=args)
    # 开始判断是路径还是文件：
    paper_list = []

    if args.paper_path.endswith(".pdf"):
        paper_list.append(Paper(path=args.paper_path))
    else:
        for root, dirs, files in os.walk(args.paper_path):
            logger.info(f"root: {root}, dirs: {dirs}, files: {files}")
            for filename in files:
                # 如果找到PDF文件，则将其复制到目标文件夹中
                if filename.endswith(".pdf"):
                    paper_list.append(Paper(path=os.path.join(root, filename)))
    logger.info(
        "------------------paper_num: {}------------------".format(len(paper_list))
    )

    for paper_index, paper_name in enumerate(paper_list):
        paper = paper_name.path.split("\\")[-1]
        logger.info(f"{paper_index}, {paper}")

    reviewer1.review_by_chatgpt(paper_list=paper_list)


def cli(args):
    parameters = ReviewerParams(**vars(args))
    main(parameters)
